@@ -62,6 +62,80 @@ class FixedSize:
 # other members
 propagated_attributes=["ptr_array", "nonnull", "chunk"]
 
+valid_attributes=set([
+    # embedded/appended at the end of the structure
+    'end',
+    # the C structure contains a pointer to data
+    # for instance we want to write an array to an allocated array
+    'to_ptr',
+    # write output to this C structure
+    'ctype',
+    # prefix for flags/values enumerations
+    'prefix',
+    # used in demarshaller to use directly data from message without a copy
+    'nocopy',
+    # store member array in a pointer
+    # similar to to_ptr but has an additional argument which is the name of a C
+    # field which will store the array length
+    'as_ptr',
+    # do not generate marshall code
+    # used for last members to be able to marshall them manually
+    'nomarshal',
+    # ??? not used by python code
+    'zero_terminated',
+    'marshall',
+    # this pointer member cannot be null
+    'nonnull',
+    # this flag member contains only a single flag
+    'unique_flag',
+    'ptr_array',
+    'outvar',
+    # C structure has an anonymous member (used in switch)
+    'anon',
+    'chunk',
+    # this channel is contained in an #ifdef section
+    # the argument specifies the preprocessor define to check
+    'ifdef',
+    # write this member as zero on network
+    'zero',
+    # specify minor version required for these members
+    'minor',
+    # this member contains the byte count for an array.
+    # the argument is the member name for item count (not bytes)
+    'bytes_count',
+    # this attribute does not exist on the network, fill just structure with the value
+    'virtual',
+    # for a switch this indicates that on network
+    # it will occupy always the same size (maximum size required for all members)
+    'fixedsize',
+])
+
+attributes_with_arguments=set([
+    'ctype',
+    'prefix',
+    'as_ptr',
+    'outvar',
+    'ifdef',
+    'minor',
+    'bytes_count',
+    'virtual',
+])
+
+def fix_attributes(attribute_list):
+    attrs = {}
+    for attr in attribute_list:
+        name = attr[0][1:]
+        lst = attr[1:]
+        if not name in valid_attributes:
+            raise Exception("Attribute %s not recognized" % name)
+        if not name in attributes_with_arguments:
+            if len(lst) > 0:
+                raise Exception("Attribute %s specified with options" % name)
+        elif len(lst) > 1:
+            raise Exception("Attribute %s has more than 1 argument" % name)
+        attrs[name] = lst
+    return attrs
+
 class Type:
     def __init__(self):
         self.attributes = {}
@@ -131,6 +205,8 @@ class Type:
         _types_by_name[self.name] = self
 
     def has_attr(self, name):
+        if not name in valid_attributes:
+            raise Exception('attribute %s not expected' % name)
         return name in self.attributes
 
 class TypeRef(Type):
@@ -178,8 +254,7 @@ class TypeAlias(Type):
         Type.__init__(self)
         self.name = name
         self.the_type = the_type
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def get_type(self, recursive=False):
         if recursive:
@@ -288,8 +363,7 @@ class EnumType(EnumBaseType):
         self.names = names
         self.values = values
 
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def __str__(self):
         return "enum %s" % self.name
@@ -342,8 +416,7 @@ class FlagsType(EnumBaseType):
         self.names = names
         self.values = values
 
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def __str__(self):
         return "flags %s" % self.name
@@ -517,6 +590,8 @@ class Containee:
         return not self.is_switch() and self.member_type.is_primitive()
 
     def has_attr(self, name):
+        if not name in valid_attributes:
+            raise Exception('attribute %s not expected' % name)
         return name in self.attributes
 
     def has_minor_attr(self):
@@ -533,15 +608,12 @@ class Member(Containee):
         Containee.__init__(self)
         self.name = name
         self.member_type = member_type
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def resolve(self, container):
         self.container = container
         self.member_type = self.member_type.resolve()
         self.member_type.register()
-        if self.has_attr("ptr32") and self.member_type.is_pointer():
-            self.member_type.set_ptr_size(4)
         for i in propagated_attributes:
             if self.has_attr(i):
                 self.member_type.attributes[i] = self.attributes[i]
@@ -636,8 +708,7 @@ class Switch(Containee):
         self.variable = variable
         self.name = name
         self.cases = cases
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def is_switch(self):
         return True
@@ -846,8 +917,7 @@ class StructType(ContainerType):
         self.members_by_name = {}
         for m in members:
             self.members_by_name[m.name] = m
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def __str__(self):
         if self.name == None:
@@ -869,8 +939,7 @@ class MessageType(ContainerType):
         for m in members:
             self.members_by_name[m.name] = m
         self.reverse_members = {} # ChannelMembers referencing this message
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def __str__(self):
         if self.name == None:
@@ -938,8 +1007,7 @@ class ChannelType(Type):
         self.base = base
         self.member_name = None
         self.members = members
-        for attr in attribute_list:
-            self.attributes[attr[0][1:]] = attr[1:]
+        self.attributes = fix_attributes(attribute_list)
 
     def __str__(self):
         if self.name == None:
@@ -1051,6 +1119,14 @@ class ProtocolType(Type):
 
         return self
 
+class FdType(IntegerType):
+    def __init__(self):
+        IntegerType.__init__(self, 0, False)
+        self.name = "fd"
+
+    def c_type(self):
+        return "int"
+
 int8 = IntegerType(8, True)
 uint8 = IntegerType(8, False)
 int16 = IntegerType(16, True)
@@ -1059,3 +1135,4 @@ int32 = IntegerType(32, True)
 uint32 = IntegerType(32, False)
 int64 = IntegerType(64, True)
 uint64 = IntegerType(64, False)
+unix_fd = FdType()
