@@ -21,15 +21,6 @@
 #include <sys/types.h>
 #endif
 
-#ifdef HAVE_SYS_SHM_H
-#include <sys/shm.h>
-#endif
-
-#ifdef HAVE_SYS_IPC_H
-#include <sys/ipc.h>
-#endif
-
-#include "glib-compat.h"
 #include "spice-client.h"
 #include "spice-common.h"
 
@@ -46,7 +37,7 @@
  * @section_id:
  * @see_also: #SpiceChannel, and the GTK widget #SpiceDisplay
  * @stability: Stable
- * @include: channel-display.h
+ * @include: spice-client.h
  *
  * A class that handles the rendering of the remote display and inform
  * of its updates.
@@ -82,6 +73,7 @@ struct _SpiceDisplayChannelPrivate {
 #ifdef G_OS_WIN32
     HDC dc;
 #endif
+    SpiceGlScanout scanout;
 };
 
 G_DEFINE_TYPE(SpiceDisplayChannel, spice_display_channel, SPICE_TYPE_CHANNEL)
@@ -92,7 +84,8 @@ enum {
     PROP_WIDTH,
     PROP_HEIGHT,
     PROP_MONITORS,
-    PROP_MONITORS_MAX
+    PROP_MONITORS_MAX,
+    PROP_GL_SCANOUT,
 };
 
 enum {
@@ -100,6 +93,7 @@ enum {
     SPICE_DISPLAY_PRIMARY_DESTROY,
     SPICE_DISPLAY_INVALIDATE,
     SPICE_DISPLAY_MARK,
+    SPICE_DISPLAY_GL_DRAW,
 
     SPICE_DISPLAY_LAST_SIGNAL,
 };
@@ -112,15 +106,37 @@ static void channel_set_handlers(SpiceChannelClass *klass);
 static void clear_surfaces(SpiceChannel *channel, gboolean keep_primary);
 static void clear_streams(SpiceChannel *channel);
 static display_surface *find_surface(SpiceDisplayChannelPrivate *c, guint32 surface_id);
-static gboolean display_stream_render(display_stream *st);
 static void spice_display_channel_reset(SpiceChannel *channel, gboolean migrating);
 static void spice_display_channel_reset_capabilities(SpiceChannel *channel);
 static void destroy_canvas(display_surface *surface);
-static void _msg_in_unref_func(gpointer data, gpointer user_data);
 static void display_session_mm_time_reset_cb(SpiceSession *session, gpointer data);
+static SpiceGlScanout* spice_gl_scanout_copy(const SpiceGlScanout *scanout);
+
+G_DEFINE_BOXED_TYPE(SpiceGlScanout, spice_gl_scanout,
+                    (GBoxedCopyFunc)spice_gl_scanout_copy,
+                    (GBoxedFreeFunc)spice_gl_scanout_free)
 
 /* ------------------------------------------------------------------ */
 
+static SpiceGlScanout*
+spice_gl_scanout_copy(const SpiceGlScanout *scanout)
+{
+    SpiceGlScanout *so = g_new(SpiceGlScanout, 1);
+
+    *so = *scanout;
+    so->fd = dup(so->fd);
+
+    return so;
+}
+
+void
+spice_gl_scanout_free(SpiceGlScanout *scanout)
+{
+    close(scanout->fd);
+
+    g_free(scanout);
+}
+
 static void spice_display_channel_dispose(GObject *object)
 {
     SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
@@ -130,6 +146,11 @@ static void spice_display_channel_dispose(GObject *object)
         c->mark_false_event_id = 0;
     }
 
+    if (c->scanout.fd >= 0) {
+        close(c->scanout.fd);
+        c->scanout.fd = -1;
+    }
+
     if (G_OBJECT_CLASS(spice_display_channel_parent_class)->dispose)
         G_OBJECT_CLASS(spice_display_channel_parent_class)->dispose(object);
 }
@@ -142,7 +163,7 @@ static void spice_display_channel_finalize(GObject *object)
     clear_surfaces(SPICE_CHANNEL(object), FALSE);
     g_hash_table_unref(c->surfaces);
     clear_streams(SPICE_CHANNEL(object));
-    g_clear_pointer(&c->palettes, cache_unref);
+    g_clear_pointer(&c->palettes, cache_free);
 
     if (G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize)
         G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize(object);
@@ -171,13 +192,13 @@ static void spice_display_channel_constructed(GObject *object)
         G_OBJECT_CLASS(spice_display_channel_parent_class)->constructed(object);
 }
 
-
 static void spice_display_get_property(GObject    *object,
                                        guint       prop_id,
                                        GValue     *value,
                                        GParamSpec *pspec)
 {
-    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
+    SpiceDisplayChannel *channel = SPICE_DISPLAY_CHANNEL(object);
+    SpiceDisplayChannelPrivate *c = channel->priv;
 
     switch (prop_id) {
     case PROP_WIDTH: {
@@ -196,6 +217,10 @@ static void spice_display_get_property(GObject    *object,
         g_value_set_uint(value, c->monitors_max);
         break;
     }
+    case PROP_GL_SCANOUT: {
+        g_value_set_static_boxed(value, spice_display_get_gl_scanout(channel));
+        break;
+    }
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
         break;
@@ -291,6 +316,23 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
                            G_PARAM_READABLE |
                            G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceDisplayChannel:gl-scanout:
+     * @display: the #SpiceDisplayChannel that emitted the signal
+     *
+     * The last #SpiceGlScanout received.
+     *
+     * Since: 0.31
+     */
+    g_object_class_install_property
+        (gobject_class, PROP_GL_SCANOUT,
+         g_param_spec_boxed("gl-scanout",
+                            "GL scanout",
+                            "GL scanout",
+                            SPICE_TYPE_GL_SCANOUT,
+                            G_PARAM_READABLE |
+                            G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceDisplayChannel::display-primary-create:
      * @display: the #SpiceDisplayChannel that emitted the signal
@@ -382,11 +424,33 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
                      1,
                      G_TYPE_INT);
 
+    /**
+     * SpiceDisplayChannel::gl-draw:
+     * @display: the #SpiceDisplayChannel that emitted the signal
+     * @x: x position
+     * @y: y position
+     * @width: width
+     * @height: height
+     *
+     * The #SpiceDisplayChannel::draw signal is emitted when the
+     * rectangular region x/y/w/h of the GL scanout is updated and
+     * must be drawn. When the draw is finished, you must call
+     * spice_display_gl_draw_done() in order to release the GL
+     * resources.
+     *
+     * Since: 0.31
+     **/
+    signals[SPICE_DISPLAY_GL_DRAW] =
+        g_signal_new("gl-draw",
+                     G_OBJECT_CLASS_TYPE(gobject_class),
+                     0, 0, NULL, NULL,
+                     g_cclosure_user_marshal_VOID__UINT_UINT_UINT_UINT,
+                     G_TYPE_NONE,
+                     4,
+                     G_TYPE_UINT, G_TYPE_UINT, G_TYPE_UINT, G_TYPE_UINT);
+
     g_type_class_add_private(klass, sizeof(SpiceDisplayChannelPrivate));
 
-    sw_canvas_init();
-    quic_init();
-    rop3_init();
     channel_set_handlers(SPICE_CHANNEL_CLASS(klass));
 }
 
@@ -419,7 +483,7 @@ gboolean spice_display_get_primary(SpiceChannel *channel, guint32 surface_id,
     primary->width = surface->width;
     primary->height = surface->height;
     primary->stride = surface->stride;
-    primary->shmid = surface->shmid;
+    primary->shmid = -1;
     primary->data = surface->data;
     primary->marked = c->mark;
     CHANNEL_DEBUG(channel, "get primary %p", primary->data);
@@ -427,6 +491,56 @@ gboolean spice_display_get_primary(SpiceChannel *channel, guint32 surface_id,
     return TRUE;
 }
 
+/**
+ * spice_display_change_preferred_compression:
+ * @channel: a #SpiceDisplayChannel
+ * @compression: a #SpiceImageCompression
+ *
+ * Tells the spice server to change the preferred image compression
+ * for the @channel.
+ *
+ * Since: 0.31
+ */
+void spice_display_change_preferred_compression(SpiceChannel *channel, gint compression)
+{
+    SpiceMsgOut *out;
+    SpiceMsgcDisplayPreferredCompression pref_comp_msg;
+
+    g_return_if_fail(SPICE_IS_DISPLAY_CHANNEL(channel));
+    g_return_if_fail(compression > SPICE_IMAGE_COMPRESSION_INVALID &&
+                     compression < SPICE_IMAGE_COMPRESSION_ENUM_END);
+
+    if (!spice_channel_test_capability(channel, SPICE_DISPLAY_CAP_PREF_COMPRESSION)) {
+        CHANNEL_DEBUG(channel, "does not have capability to change the preferred compression");
+        return;
+    }
+
+    CHANNEL_DEBUG(channel, "changing preferred compression to %d", compression);
+
+    pref_comp_msg.image_compression = compression;
+    out = spice_msg_out_new(channel, SPICE_MSGC_DISPLAY_PREFERRED_COMPRESSION);
+    out->marshallers->msgc_display_preferred_compression(out->marshaller, &pref_comp_msg);
+    spice_msg_out_send_internal(out);
+}
+
+/**
+ * spice_display_get_gl_scanout:
+ * @channel: a #SpiceDisplayChannel
+ *
+ * Retrieves the GL scanout if available
+ *
+ * Returns: the current GL scanout, or %NULL if none or not valid
+ *
+ * Since: 0.31
+ **/
+const SpiceGlScanout *
+spice_display_get_gl_scanout(SpiceDisplayChannel *channel)
+{
+    g_return_val_if_fail(SPICE_IS_DISPLAY_CHANNEL(channel), NULL);
+
+    return channel->priv->scanout.fd != -1 ? &channel->priv->scanout : NULL;
+}
+
 /* ------------------------------------------------------------------ */
 
 static void image_put(SpiceImageCache *cache, uint64_t id, pixman_image_t *image)
@@ -596,6 +710,31 @@ static void spice_display_channel_reset_capabilities(SpiceChannel *channel)
     if (SPICE_DISPLAY_CHANNEL(channel)->priv->enable_adaptive_streaming) {
         spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_STREAM_REPORT);
     }
+#ifdef G_OS_UNIX
+    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_GL_SCANOUT);
+#endif
+    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_MULTI_CODEC);
+#ifdef HAVE_BUILTIN_MJPEG
+    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_DISPLAY_CAP_CODEC_MJPEG);
+#endif
+    if (gstvideo_has_codec(SPICE_VIDEO_CODEC_TYPE_MJPEG)) {
+        spice_channel_set_capability(SPICE_CHANNEL(channel),
+                                     SPICE_DISPLAY_CAP_CODEC_MJPEG);
+    } else {
+        SPICE_DEBUG("GStreamer does not support the mjpeg codec");
+    }
+    if (gstvideo_has_codec(SPICE_VIDEO_CODEC_TYPE_VP8)) {
+        spice_channel_set_capability(SPICE_CHANNEL(channel),
+                                     SPICE_DISPLAY_CAP_CODEC_VP8);
+    } else {
+        SPICE_DEBUG("GStreamer does not support the vp8 codec");
+    }
+    if (gstvideo_has_codec(SPICE_VIDEO_CODEC_TYPE_H264)) {
+        spice_channel_set_capability(SPICE_CHANNEL(channel),
+                                     SPICE_DISPLAY_CAP_CODEC_H264);
+    } else {
+        SPICE_DEBUG("GStreamer does not support the h264 codec");
+    }
 }
 
 static void destroy_surface(gpointer data)
@@ -603,7 +742,7 @@ static void destroy_surface(gpointer data)
     display_surface *surface = data;
 
     destroy_canvas(surface);
-    g_slice_free(display_surface, surface);
+    g_free(surface);
 }
 
 static void spice_display_channel_init(SpiceDisplayChannel *channel)
@@ -620,6 +759,7 @@ static void spice_display_channel_init(SpiceDisplayChannel *channel)
     c->dc = create_compatible_dc();
 #endif
     c->monitors_max = 1;
+    c->scanout.fd = -1;
 
     if (g_getenv("SPICE_DISABLE_ADAPTIVE_STREAMING")) {
         SPICE_DEBUG("adaptive video disabled");
@@ -650,27 +790,11 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
         }
 
         CHANNEL_DEBUG(channel, "Create primary canvas");
-#if defined(WITH_X11) && defined(HAVE_SYS_SHM_H)
-        surface->shmid = shmget(IPC_PRIVATE, surface->size, IPC_CREAT | 0777);
-        if (surface->shmid >= 0) {
-            surface->data = shmat(surface->shmid, 0, 0);
-            if (surface->data == NULL) {
-                shmctl(surface->shmid, IPC_RMID, 0);
-                surface->shmid = -1;
-            }
-        }
-#else
-        surface->shmid = -1;
-#endif
-    } else {
-        surface->shmid = -1;
     }
 
-    if (surface->shmid == -1)
-        surface->data = g_malloc0(surface->size);
+    surface->data = g_malloc0(surface->size);
 
     g_return_val_if_fail(c->glz_window, 0);
-
     g_warn_if_fail(surface->canvas == NULL);
     g_warn_if_fail(surface->glz_decoder == NULL);
     g_warn_if_fail(surface->zlib_decoder == NULL);
@@ -700,7 +824,7 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
         c->primary = surface;
         g_coroutine_signal_emit(channel, signals[SPICE_DISPLAY_PRIMARY_CREATE], 0,
                                 surface->format, surface->width, surface->height,
-                                surface->stride, surface->shmid, surface->data);
+                                surface->stride, -1, surface->data);
 
         if (!spice_channel_test_capability(channel, SPICE_DISPLAY_CAP_MONITORS_CONFIG)) {
             g_array_set_size(c->monitors, 1);
@@ -724,20 +848,8 @@ static void destroy_canvas(display_surface *surface)
     zlib_decoder_destroy(surface->zlib_decoder);
     jpeg_decoder_destroy(surface->jpeg_decoder);
 
-    if (surface->shmid == -1) {
-        g_free(surface->data);
-    }
-#ifdef HAVE_SYS_SHM_H
-    else {
-        shmdt(surface->data);
-        shmctl(surface->shmid, IPC_RMID, 0);
-    }
-#endif
-    surface->shmid = -1;
-    surface->data = NULL;
-
-    surface->canvas->ops->destroy(surface->canvas);
-    surface->canvas = NULL;
+    g_clear_pointer(&surface->data, g_free);
+    g_clear_pointer(&surface->canvas, surface->canvas->ops->destroy);
 }
 
 static display_surface *find_surface(SpiceDisplayChannelPrivate *c, guint32 surface_id)
@@ -789,7 +901,6 @@ static void spice_display_channel_up(SpiceChannel *channel)
     SpiceMsgOut *out;
     SpiceSession *s = spice_channel_get_session(channel);
     SpiceMsgcDisplayInit init;
-    SpiceMsgcDisplayPreferredCompression pref_comp_msg;
     int cache_size;
     int glz_window_size;
     SpiceImageCompression preferred_compression = SPICE_IMAGE_COMPRESSION_INVALID;
@@ -812,12 +923,8 @@ static void spice_display_channel_up(SpiceChannel *channel)
     /* notify of existence of this monitor */
     g_coroutine_object_notify(G_OBJECT(channel), "monitors");
 
-    if (spice_channel_test_capability(channel, SPICE_DISPLAY_CAP_PREF_COMPRESSION) &&
-            preferred_compression > SPICE_IMAGE_COMPRESSION_INVALID) {
-        pref_comp_msg.image_compression = preferred_compression;
-        out = spice_msg_out_new(channel, SPICE_MSGC_DISPLAY_PREFERRED_COMPRESSION);
-        out->marshallers->msgc_display_preferred_compression(out->marshaller, &pref_comp_msg);
-        spice_msg_out_send_internal(out);
+    if (preferred_compression != SPICE_IMAGE_COMPRESSION_INVALID) {
+        spice_display_change_preferred_compression(channel, preferred_compression);
     }
 }
 
@@ -842,7 +949,7 @@ static void display_handle_mode(SpiceChannel *channel, SpiceMsgIn *in)
 
     g_warn_if_fail(c->mark == FALSE);
 
-    surface = g_slice_new0(display_surface);
+    surface = g_new0(display_surface, 1);
     surface->format  = mode->bits == 32 ?
         SPICE_SURFACE_FMT_32_xRGB : SPICE_SURFACE_FMT_16_555;
     surface->width   = mode->x_res;
@@ -976,7 +1083,7 @@ static void display_handle_stream_create(SpiceChannel *channel, SpiceMsgIn *in)
     SpiceMsgDisplayStreamCreate *op = spice_msg_in_parsed(in);
     display_stream *st;
 
-    CHANNEL_DEBUG(channel, "%s: id %d", __FUNCTION__, op->id);
+    CHANNEL_DEBUG(channel, "%s: id %u", __FUNCTION__, op->id);
 
     if (op->id >= c->nstreams) {
         int n = c->nstreams;
@@ -996,70 +1103,40 @@ static void display_handle_stream_create(SpiceChannel *channel, SpiceMsgIn *in)
     st->msg_create = in;
     spice_msg_in_ref(in);
     st->clip = &op->clip;
-    st->codec = op->codec_type;
     st->surface = find_surface(c, op->surface_id);
-    st->msgq = g_queue_new();
     st->channel = channel;
     st->drops_seqs_stats_arr = g_array_new(FALSE, FALSE, sizeof(drops_sequence_stats));
 
     region_init(&st->region);
     display_update_stream_region(st);
 
-    switch (st->codec) {
+    switch (op->codec_type) {
+#ifdef HAVE_BUILTIN_MJPEG
     case SPICE_VIDEO_CODEC_TYPE_MJPEG:
-        stream_mjpeg_init(st);
+        st->video_decoder = create_mjpeg_decoder(op->codec_type, st);
         break;
+#endif
+    default:
+#ifdef HAVE_GSTVIDEO
+        st->video_decoder = create_gstreamer_decoder(op->codec_type, st);
+#else
+        st->video_decoder = NULL;
+#endif
+    }
+    if (st->video_decoder == NULL) {
+        spice_printerr("could not create a video decoder for codec %u", op->codec_type);
     }
 }
 
-/* coroutine or main context */
-static gboolean display_stream_schedule(display_stream *st)
+static const SpiceRect *stream_get_dest(display_stream *st, SpiceMsgIn *frame_msg)
 {
-    SpiceSession *session = spice_channel_get_session(st->channel);
-    guint32 time, d;
-    SpiceStreamDataHeader *op;
-    SpiceMsgIn *in;
-
-    SPICE_DEBUG("%s", __FUNCTION__);
-    if (st->timeout || !session)
-        return TRUE;
-
-    time = spice_session_get_mm_time(session);
-    in = g_queue_peek_head(st->msgq);
-
-    if (in == NULL) {
-        return TRUE;
-    }
-
-    op = spice_msg_in_parsed(in);
-    if (time < op->multi_media_time) {
-        d = op->multi_media_time - time;
-        SPICE_DEBUG("scheduling next stream render in %u ms", d);
-        st->timeout = g_timeout_add(d, (GSourceFunc)display_stream_render, st);
-        return TRUE;
-    } else {
-        SPICE_DEBUG("%s: rendering too late by %u ms (ts: %u, mmtime: %u), dropping ",
-                    __FUNCTION__, time - op->multi_media_time,
-                    op->multi_media_time, time);
-        in = g_queue_pop_head(st->msgq);
-        spice_msg_in_unref(in);
-        st->num_drops_on_playback++;
-        if (g_queue_get_length(st->msgq) == 0)
-            return TRUE;
-    }
-
-    return FALSE;
-}
-
-static SpiceRect *stream_get_dest(display_stream *st)
-{
-    if (st->msg_data == NULL ||
-        spice_msg_in_type(st->msg_data) != SPICE_MSG_DISPLAY_STREAM_DATA_SIZED) {
+    if (frame_msg == NULL ||
+        spice_msg_in_type(frame_msg) != SPICE_MSG_DISPLAY_STREAM_DATA_SIZED) {
         SpiceMsgDisplayStreamCreate *info = spice_msg_in_parsed(st->msg_create);
 
         return &info->dest;
     } else {
-        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(st->msg_data);
+        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(frame_msg);
 
         return &op->dest;
    }
@@ -1074,113 +1151,70 @@ static uint32_t stream_get_flags(display_stream *st)
 }
 
 G_GNUC_INTERNAL
-uint32_t stream_get_current_frame(display_stream *st, uint8_t **data)
+uint32_t spice_msg_in_frame_data(SpiceMsgIn *frame_msg, uint8_t **data)
 {
-    if (st->msg_data == NULL) {
-        *data = NULL;
-        return 0;
+    switch (spice_msg_in_type(frame_msg)) {
+    case SPICE_MSG_DISPLAY_STREAM_DATA: {
+        SpiceMsgDisplayStreamData *op = spice_msg_in_parsed(frame_msg);
+        *data = op->data;
+        return op->data_size;
     }
-
-    if (spice_msg_in_type(st->msg_data) == SPICE_MSG_DISPLAY_STREAM_DATA) {
-        SpiceMsgDisplayStreamData *op = spice_msg_in_parsed(st->msg_data);
-
+    case SPICE_MSG_DISPLAY_STREAM_DATA_SIZED: {
+        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(frame_msg);
         *data = op->data;
         return op->data_size;
-    } else {
-        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(st->msg_data);
-
-        g_return_val_if_fail(spice_msg_in_type(st->msg_data) ==
-                             SPICE_MSG_DISPLAY_STREAM_DATA_SIZED, 0);
-        *data = op->data;
-        return op->data_size;
-   }
-
+    }
+    default:
+        *data = NULL;
+        g_return_val_if_reached(0);
+    }
 }
 
 G_GNUC_INTERNAL
-void stream_get_dimensions(display_stream *st, int *width, int *height)
+guint32 stream_get_time(display_stream *st)
 {
-    g_return_if_fail(width != NULL);
-    g_return_if_fail(height != NULL);
+    SpiceSession *session = spice_channel_get_session(st->channel);
+    return session ? spice_session_get_mm_time(session) : 0;
+}
 
-    if (st->msg_data == NULL ||
-        spice_msg_in_type(st->msg_data) != SPICE_MSG_DISPLAY_STREAM_DATA_SIZED) {
-        SpiceMsgDisplayStreamCreate *info = spice_msg_in_parsed(st->msg_create);
-
-        *width = info->stream_width;
-        *height = info->stream_height;
-    } else {
-        SpiceMsgDisplayStreamDataSized *op = spice_msg_in_parsed(st->msg_data);
-
-        *width = op->width;
-        *height = op->height;
-   }
+/* coroutine or main context */
+G_GNUC_INTERNAL
+void stream_dropped_frame_on_playback(display_stream *st)
+{
+    st->num_drops_on_playback++;
 }
 
 /* main context */
-static gboolean display_stream_render(display_stream *st)
+G_GNUC_INTERNAL
+void stream_display_frame(display_stream *st, SpiceMsgIn *frame_msg,
+                          uint32_t width, uint32_t height, uint8_t *data)
 {
-    SpiceMsgIn *in;
+    const SpiceRect *dest;
+    int stride;
 
-    st->timeout = 0;
-    do {
-        in = g_queue_pop_head(st->msgq);
+    dest = stream_get_dest(st, frame_msg);
 
-        g_return_val_if_fail(in != NULL, FALSE);
+    stride = width * sizeof(uint32_t);
+    if (!(stream_get_flags(st) & SPICE_STREAM_FLAGS_TOP_DOWN)) {
+        data += stride * (height - 1);
+        stride = -stride;
+    }
 
-        st->msg_data = in;
-        switch (st->codec) {
-        case SPICE_VIDEO_CODEC_TYPE_MJPEG:
-            stream_mjpeg_data(st);
-            break;
-        }
-
-        if (st->out_frame) {
-            int width;
-            int height;
-            SpiceRect *dest;
-            uint8_t *data;
-            int stride;
-
-            stream_get_dimensions(st, &width, &height);
-            dest = stream_get_dest(st);
-
-            data = st->out_frame;
-            stride = width * sizeof(uint32_t);
-            if (!(stream_get_flags(st) & SPICE_STREAM_FLAGS_TOP_DOWN)) {
-                data += stride * (height - 1);
-                stride = -stride;
-            }
-
-            st->surface->canvas->ops->put_image(
-                st->surface->canvas,
+    st->surface->canvas->ops->put_image(st->surface->canvas,
 #ifdef G_OS_WIN32
-                SPICE_DISPLAY_CHANNEL(st->channel)->priv->dc,
+                                        SPICE_DISPLAY_CHANNEL(st->channel)->priv->dc,
 #endif
-                dest, data,
-                width, height, stride,
-                st->have_region ? &st->region : NULL);
+                                        dest, data,
+                                        width, height, stride,
+                                        st->have_region ? &st->region : NULL);
 
-            if (st->surface->primary)
-                g_signal_emit(st->channel, signals[SPICE_DISPLAY_INVALIDATE], 0,
-                    dest->left, dest->top,
-                    dest->right - dest->left,
-                    dest->bottom - dest->top);
-        }
-
-        st->msg_data = NULL;
-        spice_msg_in_unref(in);
-
-        in = g_queue_peek_head(st->msgq);
-        if (in == NULL)
-            break;
-
-        if (display_stream_schedule(st))
-            return FALSE;
-    } while (1);
-
-    return FALSE;
+    if (st->surface->primary)
+        g_signal_emit(st->channel, signals[SPICE_DISPLAY_INVALIDATE], 0,
+                      dest->left, dest->top,
+                      dest->right - dest->left,
+                      dest->bottom - dest->top);
 }
+
 /* after a sequence of 3 drops, push a report to the server, even
  * if the report window is bigger */
 #define STREAM_REPORT_DROP_SEQ_LEN_LIMIT 3
@@ -1241,17 +1275,6 @@ static void display_update_stream_report(SpiceDisplayChannel *channel, uint32_t
     }
 }
 
-static void display_stream_reset_rendering_timer(display_stream *st)
-{
-    SPICE_DEBUG("%s", __FUNCTION__);
-    if (st->timeout != 0) {
-        g_source_remove(st->timeout);
-        st->timeout = 0;
-    }
-    while (!display_stream_schedule(st)) {
-    }
-}
-
 /*
  * Migration can occur between 2 spice-servers with different mm-times.
  * Then, the following cases can happen after migration completes:
@@ -1281,8 +1304,9 @@ static void display_stream_reset_rendering_timer(display_stream *st)
  * case 2 is less likely, since at takes at least 20 frames till the dst-server re-identifies
  * the video stream and starts sending stream data
  *
- * display_session_mm_time_reset_cb handles case 1.a, and
- * display_stream_test_frames_mm_time_reset handles case 2.b
+ * display_session_mm_time_reset_cb handles case 1.a by notifying the
+ * video decoders through their reschedule() method, and case 2.b is handled
+ * directly by the video decoders in their queue_frame() method
  */
 
 /* main context */
@@ -1300,38 +1324,9 @@ static void display_session_mm_time_reset_cb(SpiceSession *session, gpointer dat
         if (c->streams[i] == NULL) {
             continue;
         }
-        SPICE_DEBUG("%s: stream-id %d", __FUNCTION__, i);
+        SPICE_DEBUG("%s: stream-id %u", __FUNCTION__, i);
         st = c->streams[i];
-        display_stream_reset_rendering_timer(st);
-    }
-}
-
-/* coroutine context */
-static void display_stream_test_frames_mm_time_reset(display_stream *st,
-                                                     SpiceMsgIn *new_frame_msg,
-                                                     guint32 mm_time)
-{
-    SpiceStreamDataHeader *tail_op, *new_op;
-    SpiceMsgIn *tail_msg;
-
-    SPICE_DEBUG("%s", __FUNCTION__);
-    g_return_if_fail(new_frame_msg != NULL);
-    tail_msg = g_queue_peek_tail(st->msgq);
-    if (!tail_msg) {
-        return;
-    }
-    tail_op = spice_msg_in_parsed(tail_msg);
-    new_op = spice_msg_in_parsed(new_frame_msg);
-
-    if (new_op->multi_media_time < tail_op->multi_media_time) {
-        SPICE_DEBUG("new-frame-time < tail-frame-time (%u < %u):"
-                    " reseting stream, id %d",
-                    new_op->multi_media_time,
-                    tail_op->multi_media_time,
-                    new_op->id);
-        g_queue_foreach(st->msgq, _msg_in_unref_func, NULL);
-        g_queue_clear(st->msgq);
-        display_stream_reset_rendering_timer(st);
+        st->video_decoder->reschedule(st->video_decoder);
     }
 }
 
@@ -1351,10 +1346,10 @@ static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
     g_return_if_fail(c->nstreams > op->id);
 
     st =  c->streams[op->id];
-    mmtime = spice_session_get_mm_time(spice_channel_get_session(channel));
+    mmtime = stream_get_time(st);
 
     if (spice_msg_in_type(in) == SPICE_MSG_DISPLAY_STREAM_DATA_SIZED) {
-        CHANNEL_DEBUG(channel, "stream %d contains sized data", op->id);
+        CHANNEL_DEBUG(channel, "stream %u contains sized data", op->id);
     }
 
     if (op->multi_media_time == 0) {
@@ -1372,7 +1367,7 @@ static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
         CHANNEL_DEBUG(channel, "stream data too late by %u ms (ts: %u, mmtime: %u), dropping",
                       mmtime - op->multi_media_time, op->multi_media_time, mmtime);
         st->arrive_late_time += mmtime - op->multi_media_time;
-        st->num_drops_on_receive++;
+        st->arrive_late_count++;
 
         if (!st->cur_drops_seq_stats.len) {
             st->cur_drops_seq_stats.start_mm_time = op->multi_media_time;
@@ -1381,11 +1376,6 @@ static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
         st->playback_sync_drops_seq_len++;
     } else {
         CHANNEL_DEBUG(channel, "video latency: %d", latency);
-        spice_msg_in_ref(in);
-        display_stream_test_frames_mm_time_reset(st, in, mmtime);
-        g_queue_push_tail(st->msgq, in);
-        while (!display_stream_schedule(st)) {
-        }
         if (st->cur_drops_seq_stats.len) {
             st->cur_drops_seq_stats.duration = op->multi_media_time -
                                                st->cur_drops_seq_stats.start_mm_time;
@@ -1395,6 +1385,12 @@ static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
         }
         st->playback_sync_drops_seq_len = 0;
     }
+
+    /* Let the video decoder queue the frames so it can optimize their
+     * decoding and best decide if/when to drop them when they are late,
+     * taking into account the impact on later frames.
+     */
+    st->video_decoder->queue_frame(st->video_decoder, in,  latency);
     if (c->enable_adaptive_streaming) {
         display_update_stream_report(SPICE_DISPLAY_CHANNEL(channel), op->id,
                                      op->multi_media_time, latency);
@@ -1427,11 +1423,6 @@ static void display_handle_stream_clip(SpiceChannel *channel, SpiceMsgIn *in)
     display_update_stream_region(st);
 }
 
-static void _msg_in_unref_func(gpointer data, gpointer user_data)
-{
-    spice_msg_in_unref(data);
-}
-
 static void destroy_stream(SpiceChannel *channel, int id)
 {
     SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
@@ -1448,15 +1439,15 @@ static void destroy_stream(SpiceChannel *channel, int id)
     if (!st)
         return;
 
-    num_out_frames = st->num_input_frames - st->num_drops_on_receive - st->num_drops_on_playback;
-    CHANNEL_DEBUG(channel, "%s: id=%d #in-frames=%d out/in=%.2f "
-        "#drops-on-receive=%d avg-late-time(ms)=%.2f "
-        "#drops-on-playback=%d", __FUNCTION__,
+    num_out_frames = st->num_input_frames - st->arrive_late_count - st->num_drops_on_playback;
+    CHANNEL_DEBUG(channel, "%s: id=%d #in-frames=%u out/in=%.2f "
+        "#drops-on-receive=%u avg-late-time(ms)=%.2f "
+        "#drops-on-playback=%u", __FUNCTION__,
         id,
         st->num_input_frames,
         num_out_frames / (double)st->num_input_frames,
-        st->num_drops_on_receive,
-        st->num_drops_on_receive ? st->arrive_late_time / ((double)st->num_drops_on_receive): 0,
+        st->arrive_late_count,
+        st->arrive_late_count ? st->arrive_late_time / ((double)st->arrive_late_count): 0,
         st->num_drops_on_playback);
     if (st->num_drops_seqs) {
         CHANNEL_DEBUG(channel, "%s: #drops-sequences=%u ==>", __FUNCTION__, st->num_drops_seqs);
@@ -1477,20 +1468,14 @@ static void destroy_stream(SpiceChannel *channel, int id)
 
     g_array_free(st->drops_seqs_stats_arr, TRUE);
 
-    switch (st->codec) {
-    case SPICE_VIDEO_CODEC_TYPE_MJPEG:
-        stream_mjpeg_cleanup(st);
-        break;
+    if (st->video_decoder) {
+        st->video_decoder->destroy(st->video_decoder);
     }
 
     if (st->msg_clip)
         spice_msg_in_unref(st->msg_clip);
     spice_msg_in_unref(st->msg_create);
 
-    g_queue_foreach(st->msgq, _msg_in_unref_func, NULL);
-    g_queue_free(st->msgq);
-    if (st->timeout != 0)
-        g_source_remove(st->timeout);
     g_free(st);
     c->streams[id] = NULL;
 }
@@ -1503,8 +1488,7 @@ static void clear_streams(SpiceChannel *channel)
     for (i = 0; i < c->nstreams; i++) {
         destroy_stream(channel, i);
     }
-    g_free(c->streams);
-    c->streams = NULL;
+    g_clear_pointer(&c->streams, g_free);
     c->nstreams = 0;
 }
 
@@ -1514,7 +1498,7 @@ static void display_handle_stream_destroy(SpiceChannel *channel, SpiceMsgIn *in)
     SpiceMsgDisplayStreamDestroy *op = spice_msg_in_parsed(in);
 
     g_return_if_fail(op != NULL);
-    CHANNEL_DEBUG(channel, "%s: id %d", __FUNCTION__, op->id);
+    CHANNEL_DEBUG(channel, "%s: id %u", __FUNCTION__, op->id);
     destroy_stream(channel, op->id);
 }
 
@@ -1646,7 +1630,7 @@ static void display_handle_surface_create(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgSurfaceCreate *create = spice_msg_in_parsed(in);
-    display_surface *surface = g_slice_new0(display_surface);
+    display_surface *surface = g_new0(display_surface, 1);
 
     surface->surface_id = create->surface_id;
     surface->format = create->format;
@@ -1656,7 +1640,7 @@ static void display_handle_surface_create(SpiceChannel *channel, SpiceMsgIn *in)
     surface->size   = surface->height * surface->stride;
 
     if (create->flags & SPICE_SURFACE_FLAGS_PRIMARY) {
-        SPICE_DEBUG("primary flags: %d", create->flags);
+        SPICE_DEBUG("primary flags: %x", create->flags);
         surface->primary = true;
         create_canvas(channel, surface);
         if (c->mark_false_event_id != 0) {
@@ -1720,7 +1704,11 @@ static void display_handle_monitors_config(SpiceChannel *channel, SpiceMsgIn *in
     guint i;
 
     g_return_if_fail(config != NULL);
-    g_return_if_fail(config->count > 0);
+
+    if (config->count == 0) {
+        CHANNEL_DEBUG(channel, "received empty monitor config");
+        return;
+    }
 
     CHANNEL_DEBUG(channel, "received new monitors config from guest: n: %d/%d", config->count, config->max_allowed);
 
@@ -1754,6 +1742,71 @@ static void display_handle_monitors_config(SpiceChannel *channel, SpiceMsgIn *in
     g_coroutine_object_notify(G_OBJECT(channel), "monitors");
 }
 
+
+#ifdef G_OS_UNIX
+/* coroutine context */
+static void display_handle_gl_scanout_unix(SpiceChannel *channel, SpiceMsgIn *in)
+{
+    SpiceDisplayChannel *display = SPICE_DISPLAY_CHANNEL(channel);
+    SpiceDisplayChannelPrivate *c = display->priv;
+    SpiceMsgDisplayGlScanoutUnix *scanout = spice_msg_in_parsed(in);
+
+    scanout->drm_dma_buf_fd = -1;
+    if (scanout->drm_fourcc_format != 0) {
+        scanout->drm_dma_buf_fd = spice_channel_unix_read_fd(channel);
+        CHANNEL_DEBUG(channel, "gl scanout fd: %d", scanout->drm_dma_buf_fd);
+    }
+
+    c->scanout.y0top = scanout->flags & SPICE_GL_SCANOUT_FLAGS_Y0TOP;
+    if (c->scanout.fd >= 0)
+        close(c->scanout.fd);
+    c->scanout.fd = scanout->drm_dma_buf_fd;
+    c->scanout.width = scanout->width;
+    c->scanout.height = scanout->height;
+    c->scanout.stride = scanout->stride;
+    c->scanout.format = scanout->drm_fourcc_format;
+
+    g_coroutine_object_notify(G_OBJECT(channel), "gl-scanout");
+}
+#endif
+
+/* coroutine context */
+static void display_handle_gl_draw(SpiceChannel *channel, SpiceMsgIn *in)
+{
+    SpiceMsgDisplayGlDraw *draw = spice_msg_in_parsed(in);
+
+    CHANNEL_DEBUG(channel, "gl draw %ux%u+%u+%u",
+                  draw->w, draw->h, draw->x, draw->y);
+
+    g_coroutine_signal_emit(channel, signals[SPICE_DISPLAY_GL_DRAW], 0,
+                            draw->x, draw->y,
+                            draw->w, draw->h);
+}
+
+/**
+ * spice_display_gl_draw_done:
+ * @channel: a #SpiceDisplayChannel
+ *
+ * After a SpiceDisplayChannel::gl-draw is emitted, the client should
+ * draw the current display with the current GL scanout, and must
+ * release the GL resource with a call to spice_display_gl_draw_done()
+ * (failing to do so for each gl-draw may result in a frozen display).
+ *
+ * Since: 0.31
+ **/
+void spice_display_gl_draw_done(SpiceDisplayChannel *display)
+{
+    SpiceChannel *channel;
+    SpiceMsgOut *out;
+
+    g_return_if_fail(SPICE_IS_DISPLAY_CHANNEL(display));
+    channel = SPICE_CHANNEL(display);
+
+    out = spice_msg_out_new(channel, SPICE_MSGC_DISPLAY_GL_DRAW_DONE);
+    out->marshallers->msgc_display_gl_draw_done(out->marshaller, NULL);
+    spice_msg_out_send_internal(out);
+}
+
 static void channel_set_handlers(SpiceChannelClass *klass)
 {
     static const spice_msg_handler handlers[] = {
@@ -1792,6 +1845,10 @@ static void channel_set_handlers(SpiceChannelClass *klass)
         [ SPICE_MSG_DISPLAY_SURFACE_DESTROY ]    = display_handle_surface_destroy,
 
         [ SPICE_MSG_DISPLAY_MONITORS_CONFIG ]    = display_handle_monitors_config,
+#ifdef G_OS_UNIX
+        [ SPICE_MSG_DISPLAY_GL_SCANOUT_UNIX ]    = display_handle_gl_scanout_unix,
+#endif
+        [ SPICE_MSG_DISPLAY_GL_DRAW ]            = display_handle_gl_draw,
     };
 
     spice_channel_set_handlers(klass, handlers, G_N_ELEMENTS(handlers));
