@@ -24,7 +24,7 @@
 #ifdef USE_USBREDIR
 #include <glib/gi18n.h>
 #include <usbredirhost.h>
-#if USE_POLKIT
+#ifdef USE_POLKIT
 #include "usb-acl-helper.h"
 #endif
 #include "channel-usbredir-priv.h"
@@ -36,15 +36,14 @@
 #include "spice-common.h"
 
 #include "spice-channel-priv.h"
-#include "glib-compat.h"
 
 /**
  * SECTION:channel-usbredir
  * @short_description: usb redirection
  * @title: USB Redirection Channel
  * @section_id:
- * @stability: API Stable (channel in development)
- * @include: channel-usbredir.h
+ * @stability: Stable
+ * @include: spice-client.h
  *
  * The Spice protocol defines a set of messages to redirect USB devices
  * from the Spice client to the VM. This channel handles these messages.
@@ -57,7 +56,7 @@
 
 enum SpiceUsbredirChannelState {
     STATE_DISCONNECTED,
-#if USE_POLKIT
+#ifdef USE_POLKIT
     STATE_WAITING_FOR_ACL_HELPER,
 #endif
     STATE_CONNECTED,
@@ -75,10 +74,11 @@ struct _SpiceUsbredirChannelPrivate {
     const uint8_t *read_buf;
     int read_buf_size;
     enum SpiceUsbredirChannelState state;
-#if USE_POLKIT
-    GSimpleAsyncResult *result;
+#ifdef USE_POLKIT
+    GTask *task;
     SpiceUsbAclHelper *acl_helper;
 #endif
+    GMutex device_connect_mutex;
 };
 
 static void channel_set_handlers(SpiceChannelClass *klass);
@@ -91,6 +91,9 @@ static void usbredir_log(void *user_data, int level, const char *msg);
 static int usbredir_read_callback(void *user_data, uint8_t *data, int count);
 static int usbredir_write_callback(void *user_data, uint8_t *data, int count);
 static void usbredir_write_flush_callback(void *user_data);
+#if USBREDIR_VERSION >= 0x000701
+static uint64_t usbredir_buffered_output_size_callback(void *user_data);
+#endif
 
 static void *usbredir_alloc_lock(void);
 static void usbredir_lock_lock(void *user_data);
@@ -107,24 +110,59 @@ static void spice_usbredir_channel_init(SpiceUsbredirChannel *channel)
 {
 #ifdef USE_USBREDIR
     channel->priv = SPICE_USBREDIR_CHANNEL_GET_PRIVATE(channel);
+    g_mutex_init(&channel->priv->device_connect_mutex);
 #endif
 }
 
 #ifdef USE_USBREDIR
+
+static void _channel_reset_finish(SpiceUsbredirChannel *channel)
+{
+    SpiceUsbredirChannelPrivate *priv = channel->priv;
+
+    spice_usbredir_channel_lock(channel);
+
+    usbredirhost_close(priv->host);
+    priv->host = NULL;
+
+    /* Call set_context to re-create the host */
+    spice_usbredir_channel_set_context(channel, priv->context);
+
+    spice_usbredir_channel_unlock(channel);
+}
+
+static void _channel_reset_cb(GObject *gobject,
+                              GAsyncResult *result,
+                              gpointer user_data)
+{
+    SpiceChannel *spice_channel =  SPICE_CHANNEL(gobject);
+    SpiceUsbredirChannel *channel = SPICE_USBREDIR_CHANNEL(spice_channel);
+    gboolean migrating = GPOINTER_TO_UINT(user_data);
+    GError *err = NULL;
+
+    _channel_reset_finish(channel);
+
+    SPICE_CHANNEL_CLASS(spice_usbredir_channel_parent_class)->channel_reset(spice_channel, migrating);
+
+    spice_usbredir_channel_disconnect_device_finish(channel, result, &err);
+    g_object_unref(result);
+}
+
 static void spice_usbredir_channel_reset(SpiceChannel *c, gboolean migrating)
 {
     SpiceUsbredirChannel *channel = SPICE_USBREDIR_CHANNEL(c);
     SpiceUsbredirChannelPrivate *priv = channel->priv;
 
     if (priv->host) {
-        if (priv->state == STATE_CONNECTED)
-            spice_usbredir_channel_disconnect_device(channel);
-        usbredirhost_close(priv->host);
-        priv->host = NULL;
-        /* Call set_context to re-create the host */
-        spice_usbredir_channel_set_context(channel, priv->context);
+        if (priv->state == STATE_CONNECTED) {
+            spice_usbredir_channel_disconnect_device_async(channel, NULL,
+                _channel_reset_cb, GUINT_TO_POINTER(migrating));
+        } else {
+            _channel_reset_finish(channel);
+        }
+    } else {
+        SPICE_CHANNEL_CLASS(spice_usbredir_channel_parent_class)->channel_reset(c, migrating);
     }
-    SPICE_CHANNEL_CLASS(spice_usbredir_channel_parent_class)->channel_reset(c, migrating);
 }
 #endif
 
@@ -182,6 +220,9 @@ static void spice_usbredir_channel_finalize(GObject *obj)
 
     if (channel->priv->host)
         usbredirhost_close(channel->priv->host);
+#ifdef USE_USBREDIR
+    g_mutex_clear(&channel->priv->device_connect_mutex);
+#endif
 
     /* Chain up to the parent class */
     if (G_OBJECT_CLASS(spice_usbredir_channel_parent_class)->finalize)
@@ -224,6 +265,10 @@ void spice_usbredir_channel_set_context(SpiceUsbredirChannel *channel,
                                    usbredirhost_fl_write_cb_owns_buffer);
     if (!priv->host)
         g_error("Out of memory allocating usbredirhost");
+
+#if USBREDIR_VERSION >= 0x000701
+    usbredirhost_set_buffered_output_size_cb(priv->host, usbredir_buffered_output_size_callback);
+#endif
 }
 
 static gboolean spice_usbredir_channel_open_device(
@@ -234,7 +279,7 @@ static gboolean spice_usbredir_channel_open_device(
     int rc, status;
 
     g_return_val_if_fail(priv->state == STATE_DISCONNECTED
-#if USE_POLKIT
+#ifdef USE_POLKIT
                          || priv->state == STATE_WAITING_FOR_ACL_HELPER
 #endif
                          , FALSE);
@@ -268,7 +313,7 @@ static gboolean spice_usbredir_channel_open_device(
     return TRUE;
 }
 
-#if USE_POLKIT
+#ifdef USE_POLKIT
 static void spice_usbredir_channel_open_acl_cb(
     GObject *gobject, GAsyncResult *acl_res, gpointer user_data)
 {
@@ -290,21 +335,49 @@ static void spice_usbredir_channel_open_acl_cb(
         spice_usbredir_channel_open_device(channel, &err);
     }
     if (err) {
-        g_simple_async_result_take_error(priv->result, err);
-        libusb_unref_device(priv->device);
-        priv->device = NULL;
+        g_clear_pointer(&priv->device, libusb_unref_device);
         g_boxed_free(spice_usb_device_get_type(), priv->spice_device);
         priv->spice_device = NULL;
         priv->state  = STATE_DISCONNECTED;
+        g_task_return_error(priv->task, err);
+    } else {
+        g_task_return_boolean(priv->task, TRUE);
     }
 
-    spice_usb_acl_helper_close_acl(priv->acl_helper);
     g_clear_object(&priv->acl_helper);
     g_object_set(spice_channel_get_session(SPICE_CHANNEL(channel)),
                  "inhibit-keyboard-grab", FALSE, NULL);
 
-    g_simple_async_result_complete_in_idle(priv->result);
-    g_clear_object(&priv->result);
+    g_clear_object(&priv->task);
+}
+#endif
+
+#ifndef USE_POLKIT
+static void
+_open_device_async_cb(GTask *task,
+                      gpointer object,
+                      gpointer task_data,
+                      GCancellable *cancellable)
+{
+    GError *err = NULL;
+    SpiceUsbredirChannel *channel = SPICE_USBREDIR_CHANNEL(object);
+    SpiceUsbredirChannelPrivate *priv = channel->priv;
+
+    spice_usbredir_channel_lock(channel);
+
+    if (!spice_usbredir_channel_open_device(channel, &err)) {
+        g_clear_pointer(&priv->device, libusb_unref_device);
+        g_boxed_free(spice_usb_device_get_type(), priv->spice_device);
+        priv->spice_device = NULL;
+    }
+
+    spice_usbredir_channel_unlock(channel);
+
+    if (err) {
+        g_task_return_error(task, err);
+    } else {
+        g_task_return_boolean(task, TRUE);
+    }
 }
 #endif
 
@@ -318,10 +391,7 @@ void spice_usbredir_channel_connect_device_async(
                                           gpointer              user_data)
 {
     SpiceUsbredirChannelPrivate *priv = channel->priv;
-    GSimpleAsyncResult *result;
-#if ! USE_POLKIT
-    GError *err = NULL;
-#endif
+    GTask *task;
 
     g_return_if_fail(SPICE_IS_USBREDIR_CHANNEL(channel));
     g_return_if_fail(device != NULL);
@@ -331,18 +401,17 @@ void spice_usbredir_channel_connect_device_async(
                   spice_usb_device_get_pid(spice_device),
                   spice_device, channel);
 
-    result = g_simple_async_result_new(G_OBJECT(channel), callback, user_data,
-                                 spice_usbredir_channel_connect_device_async);
+    task = g_task_new(channel, cancellable, callback, user_data);
 
     if (!priv->host) {
-        g_simple_async_result_set_error(result,
+        g_task_return_new_error(task,
                             SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
                             "Error libusb context not set");
         goto done;
     }
 
     if (priv->state != STATE_DISCONNECTED) {
-        g_simple_async_result_set_error(result,
+        g_task_return_new_error(task,
                             SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
                             "Error channel is busy");
         goto done;
@@ -351,32 +420,25 @@ void spice_usbredir_channel_connect_device_async(
     priv->device = libusb_ref_device(device);
     priv->spice_device = g_boxed_copy(spice_usb_device_get_type(),
                                       spice_device);
-#if USE_POLKIT
-    priv->result = result;
+#ifdef USE_POLKIT
+    priv->task = task;
     priv->state  = STATE_WAITING_FOR_ACL_HELPER;
     priv->acl_helper = spice_usb_acl_helper_new();
     g_object_set(spice_channel_get_session(SPICE_CHANNEL(channel)),
                  "inhibit-keyboard-grab", TRUE, NULL);
-    spice_usb_acl_helper_open_acl(priv->acl_helper,
-                                  libusb_get_bus_number(device),
-                                  libusb_get_device_address(device),
-                                  cancellable,
-                                  spice_usbredir_channel_open_acl_cb,
-                                  channel);
+    spice_usb_acl_helper_open_acl_async(priv->acl_helper,
+                                        libusb_get_bus_number(device),
+                                        libusb_get_device_address(device),
+                                        cancellable,
+                                        spice_usbredir_channel_open_acl_cb,
+                                        channel);
     return;
 #else
-    if (!spice_usbredir_channel_open_device(channel, &err)) {
-        g_simple_async_result_take_error(result, err);
-        libusb_unref_device(priv->device);
-        priv->device = NULL;
-        g_boxed_free(spice_usb_device_get_type(), priv->spice_device);
-        priv->spice_device = NULL;
-    }
+    g_task_run_in_thread(task, _open_device_async_cb);
 #endif
 
 done:
-    g_simple_async_result_complete_in_idle(result);
-    g_object_unref(result);
+    g_object_unref(task);
 }
 
 G_GNUC_INTERNAL
@@ -385,16 +447,11 @@ gboolean spice_usbredir_channel_connect_device_finish(
                                                GAsyncResult         *res,
                                                GError              **err)
 {
-    GSimpleAsyncResult *result = G_SIMPLE_ASYNC_RESULT(res);
+    GTask *task = G_TASK(res);
 
-    g_return_val_if_fail(g_simple_async_result_is_valid(res, G_OBJECT(channel),
-                                 spice_usbredir_channel_connect_device_async),
-                         FALSE);
+    g_return_val_if_fail(g_task_is_valid(task, channel), FALSE);
 
-    if (g_simple_async_result_propagate_error(result, err))
-        return FALSE;
-
-    return TRUE;
+    return g_task_propagate_boolean(task, err);
 }
 
 G_GNUC_INTERNAL
@@ -404,15 +461,17 @@ void spice_usbredir_channel_disconnect_device(SpiceUsbredirChannel *channel)
 
     CHANNEL_DEBUG(channel, "disconnecting device from usb channel %p", channel);
 
+    spice_usbredir_channel_lock(channel);
+
     switch (priv->state) {
     case STATE_DISCONNECTED:
     case STATE_DISCONNECTING:
         break;
-#if USE_POLKIT
+#ifdef USE_POLKIT
     case STATE_WAITING_FOR_ACL_HELPER:
         priv->state = STATE_DISCONNECTING;
         /* We're still waiting for the acl helper -> cancel it */
-        spice_usb_acl_helper_close_acl(priv->acl_helper);
+        spice_usb_acl_helper_cancel(priv->acl_helper);
         break;
 #endif
     case STATE_CONNECTED:
@@ -430,13 +489,44 @@ void spice_usbredir_channel_disconnect_device(SpiceUsbredirChannel *channel)
         }
         /* This also closes the libusb handle we passed from open_device */
         usbredirhost_set_device(priv->host, NULL);
-        libusb_unref_device(priv->device);
-        priv->device = NULL;
+        g_clear_pointer(&priv->device, libusb_unref_device);
         g_boxed_free(spice_usb_device_get_type(), priv->spice_device);
         priv->spice_device = NULL;
         priv->state  = STATE_DISCONNECTED;
         break;
     }
+
+    spice_usbredir_channel_unlock(channel);
+}
+
+static void
+_disconnect_device_thread(GTask *task,
+                          gpointer object,
+                          gpointer task_data,
+                          GCancellable *cancellable)
+{
+    spice_usbredir_channel_disconnect_device(SPICE_USBREDIR_CHANNEL(object));
+    g_task_return_boolean(task, TRUE);
+}
+
+G_GNUC_INTERNAL
+gboolean spice_usbredir_channel_disconnect_device_finish(
+                                               SpiceUsbredirChannel *channel,
+                                               GAsyncResult         *res,
+                                               GError              **err)
+{
+    return g_task_propagate_boolean(G_TASK(res), err);
+}
+
+void spice_usbredir_channel_disconnect_device_async(SpiceUsbredirChannel *channel,
+                                                    GCancellable *cancellable,
+                                                    GAsyncReadyCallback callback,
+                                                    gpointer user_data)
+{
+    GTask* task = g_task_new(channel, cancellable, callback, user_data);
+
+    g_return_if_fail(channel != NULL);
+    g_task_run_in_thread(task, _disconnect_device_thread);
 }
 
 G_GNUC_INTERNAL
@@ -461,6 +551,14 @@ void spice_usbredir_channel_get_guest_filter(
 /* ------------------------------------------------------------------ */
 /* callbacks (any context)                                            */
 
+#if USBREDIR_VERSION >= 0x000701
+static uint64_t usbredir_buffered_output_size_callback(void *user_data)
+{
+    g_return_val_if_fail(SPICE_IS_USBREDIR_CHANNEL(user_data), 0);
+    return spice_channel_get_queue_size(SPICE_CHANNEL(user_data));
+}
+#endif
+
 /* Note that this function must be re-entrant safe, as it can get called
    from both the main thread as well as from the usb event handling thread */
 static void usbredir_write_flush_callback(void *user_data)
@@ -549,16 +647,22 @@ static int usbredir_write_callback(void *user_data, uint8_t *data, int count)
 }
 
 static void *usbredir_alloc_lock(void) {
-#if GLIB_CHECK_VERSION(2,32,0)
     GMutex *mutex;
 
     mutex = g_new0(GMutex, 1);
     g_mutex_init(mutex);
 
     return mutex;
-#else
-    return g_mutex_new();
-#endif
+}
+
+void spice_usbredir_channel_lock(SpiceUsbredirChannel *channel)
+{
+    g_mutex_lock(&channel->priv->device_connect_mutex);
+}
+
+void spice_usbredir_channel_unlock(SpiceUsbredirChannel *channel)
+{
+    g_mutex_unlock(&channel->priv->device_connect_mutex);
 }
 
 static void usbredir_lock_lock(void *user_data) {
@@ -576,12 +680,8 @@ static void usbredir_unlock_lock(void *user_data) {
 static void usbredir_free_lock(void *user_data) {
     GMutex *mutex = user_data;
 
-#if GLIB_CHECK_VERSION(2,32,0)
     g_mutex_clear(mutex);
     g_free(mutex);
-#else
-    g_mutex_free(mutex);
-#endif
 }
 
 /* --------------------------------------------------------------------- */
@@ -628,7 +728,6 @@ static void usbredir_handle_msg(SpiceChannel *c, SpiceMsgIn *in)
 {
     SpiceUsbredirChannel *channel = SPICE_USBREDIR_CHANNEL(c);
     SpiceUsbredirChannelPrivate *priv = channel->priv;
-    device_error_data data;
     int r, size;
     uint8_t *buf;
 
@@ -641,13 +740,19 @@ static void usbredir_handle_msg(SpiceChannel *c, SpiceMsgIn *in)
     priv->read_buf = buf;
     priv->read_buf_size = size;
 
+    spice_usbredir_channel_lock(channel);
+
     r = usbredirhost_read_guest_data(priv->host);
     if (r != 0) {
         SpiceUsbDevice *spice_device = priv->spice_device;
+        device_error_data err_data;
         gchar *desc;
         GError *err;
 
-        g_return_if_fail(spice_device != NULL);
+        if (spice_device == NULL) {
+            spice_usbredir_channel_unlock(channel);
+            return;
+        }
 
         desc = spice_usb_device_get_description(spice_device, NULL);
         switch (r) {
@@ -673,17 +778,19 @@ static void usbredir_handle_msg(SpiceChannel *c, SpiceMsgIn *in)
 
         CHANNEL_DEBUG(c, "%s", err->message);
 
-        data.channel = channel;
-        data.caller = coroutine_self();
-        data.spice_device = g_boxed_copy(spice_usb_device_get_type(), spice_device);
-        data.error = err;
-        g_idle_add(device_error, &data);
+        err_data.channel = channel;
+        err_data.caller = coroutine_self();
+        err_data.spice_device = g_boxed_copy(spice_usb_device_get_type(), spice_device);
+        err_data.error = err;
+        g_idle_add(device_error, &err_data);
         coroutine_yield(NULL);
 
-        g_boxed_free(spice_usb_device_get_type(), data.spice_device);
+        g_boxed_free(spice_usb_device_get_type(), err_data.spice_device);
 
         g_error_free(err);
     }
+
+    spice_usbredir_channel_unlock(channel);
 }
 
 #endif /* USE_USBREDIR */
