@@ -22,7 +22,6 @@
 #include "spice-channel-priv.h"
 #include "spice-session-priv.h"
 #include "spice-marshal.h"
-#include "glib-compat.h"
 #include "vmcstream.h"
 #include "giopipe.h"
 
@@ -33,7 +32,7 @@
  * @section_id:
  * @see_also: #SpiceChannel
  * @stability: Stable
- * @include: channel-webdav.h
+ * @include: spice-client.h
  *
  * The "webdav" channel exports a directory to the guest for file
  * manipulation (read/write/copy etc). The underlying protocol is
