@@ -19,7 +19,6 @@
 
 #include "spice-client.h"
 #include "spice-common.h"
-#include "glib-compat.h"
 
 #include "spice-channel-priv.h"
 #include "spice-session-priv.h"
@@ -63,7 +62,7 @@ static gboolean channel_connect(SpiceChannel *channel, gboolean tls);
  * @section_id:
  * @see_also: #SpiceSession, #SpiceMainChannel and other channels
  * @stability: Stable
- * @include: spice-channel.h
+ * @include: spice-client.h
  *
  * #SpiceChannel is the base class for the different kind of Spice
  * channel connections, such as #SpiceMainChannel, or
@@ -110,7 +109,8 @@ static void spice_channel_init(SpiceChannel *channel)
     c->out_serial = 1;
     c->in_serial = 1;
     c->fd = -1;
-    c->auth_needs_username_and_password = FALSE;
+    c->auth_needs_username = FALSE;
+    c->auth_needs_password = FALSE;
     strcpy(c->name, "?");
     c->caps = g_array_new(FALSE, TRUE, sizeof(guint32));
     c->common_caps = g_array_new(FALSE, TRUE, sizeof(guint32));
@@ -118,11 +118,11 @@ static void spice_channel_init(SpiceChannel *channel)
     c->remote_common_caps = g_array_new(FALSE, TRUE, sizeof(guint32));
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_PROTOCOL_AUTH_SELECTION);
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_MINI_HEADER);
-#if HAVE_SASL
+#ifdef HAVE_SASL
     spice_channel_set_common_capability(channel, SPICE_COMMON_CAP_AUTH_SASL);
 #endif
     g_queue_init(&c->xmit_queue);
-    STATIC_MUTEX_INIT(c->xmit_queue_lock);
+    g_mutex_init(&c->xmit_queue_lock);
 }
 
 static void spice_channel_constructed(GObject *gobject)
@@ -155,10 +155,7 @@ static void spice_channel_dispose(GObject *gobject)
 
     spice_channel_disconnect(channel, SPICE_CHANNEL_CLOSED);
 
-    if (c->session) {
-         g_object_unref(c->session);
-         c->session = NULL;
-    }
+    g_clear_object(&c->session);
 
     g_clear_error(&c->error);
 
@@ -176,7 +173,7 @@ static void spice_channel_finalize(GObject *gobject)
 
     g_idle_remove_by_data(gobject);
 
-    STATIC_MUTEX_CLEAR(c->xmit_queue_lock);
+    g_mutex_clear(&c->xmit_queue_lock);
 
     if (c->caps)
         g_array_free(c->caps, TRUE);
@@ -190,6 +187,8 @@ static void spice_channel_finalize(GObject *gobject)
     if (c->remote_common_caps)
         g_array_free(c->remote_common_caps, TRUE);
 
+    g_clear_pointer(&c->peer_msg, g_free);
+
     /* Chain up to the parent class */
     if (G_OBJECT_CLASS(spice_channel_parent_class)->finalize)
         G_OBJECT_CLASS(spice_channel_parent_class)->finalize(gobject);
@@ -372,9 +371,9 @@ static inline void spice_header_set_msg_type(uint8_t *header, gboolean is_mini_h
                                              uint16_t type)
 {
     if (is_mini_header) {
-        ((SpiceMiniDataHeader *)header)->type = type;
+        ((SpiceMiniDataHeader *)header)->type = GUINT16_TO_LE(type);
     } else {
-        ((SpiceDataHeader *)header)->type = type;
+        ((SpiceDataHeader *)header)->type = GUINT16_TO_LE(type);
     }
 }
 
@@ -382,9 +381,9 @@ static inline void spice_header_set_msg_size(uint8_t *header, gboolean is_mini_h
                                              uint32_t size)
 {
     if (is_mini_header) {
-        ((SpiceMiniDataHeader *)header)->size = size;
+        ((SpiceMiniDataHeader *)header)->size = GUINT32_TO_LE(size);
     } else {
-        ((SpiceDataHeader *)header)->size = size;
+        ((SpiceDataHeader *)header)->size = GUINT32_TO_LE(size);
     }
 }
 
@@ -392,9 +391,9 @@ G_GNUC_INTERNAL
 uint16_t spice_header_get_msg_type(uint8_t *header, gboolean is_mini_header)
 {
     if (is_mini_header) {
-        return ((SpiceMiniDataHeader *)header)->type;
+        return GUINT16_FROM_LE(((SpiceMiniDataHeader *)header)->type);
     } else {
-        return ((SpiceDataHeader *)header)->type;
+        return GUINT16_FROM_LE(((SpiceDataHeader *)header)->type);
     }
 }
 
@@ -402,9 +401,9 @@ G_GNUC_INTERNAL
 uint32_t spice_header_get_msg_size(uint8_t *header, gboolean is_mini_header)
 {
     if (is_mini_header) {
-        return ((SpiceMiniDataHeader *)header)->size;
+        return GUINT32_FROM_LE(((SpiceMiniDataHeader *)header)->size);
     } else {
-        return ((SpiceDataHeader *)header)->size;
+        return GUINT32_FROM_LE(((SpiceDataHeader *)header)->size);
     }
 }
 
@@ -417,7 +416,7 @@ static inline void spice_header_set_msg_serial(uint8_t *header, gboolean is_mini
                                                uint64_t serial)
 {
     if (!is_mini_header) {
-        ((SpiceDataHeader *)header)->serial = serial;
+        ((SpiceDataHeader *)header)->serial = GUINT64_TO_LE(serial);
     }
 }
 
@@ -456,7 +455,7 @@ static inline uint32_t spice_header_get_msg_sub_list(uint8_t *header, gboolean i
     if (is_mini_header) {
         return 0;
     } else {
-        return ((SpiceDataHeader *)header)->sub_list;
+        return GUINT32_FROM_LE(((SpiceDataHeader *)header)->sub_list);
     }
 }
 
@@ -470,7 +469,7 @@ SpiceMsgIn *spice_msg_in_new(SpiceChannel *channel)
 
     g_return_val_if_fail(channel != NULL, NULL);
 
-    in = g_slice_new0(SpiceMsgIn);
+    in = g_new0(SpiceMsgIn, 1);
     in->refcount = 1;
     in->channel  = channel;
 
@@ -518,7 +517,7 @@ void spice_msg_in_unref(SpiceMsgIn *in)
     } else {
         g_free(in->data);
     }
-    g_slice_free(SpiceMsgIn, in);
+    g_free(in);
 }
 
 G_GNUC_INTERNAL
@@ -569,7 +568,7 @@ void spice_msg_in_hexdump(SpiceMsgIn *in)
 {
     SpiceChannelPrivate *c = in->channel->priv;
 
-    fprintf(stderr, "--\n<< hdr: %s serial %" PRIu64 " type %d size %d sub-list %d\n",
+    fprintf(stderr, "--\n<< hdr: %s serial %" PRIu64 " type %u size %u sub-list %u\n",
             c->name, spice_header_get_in_msg_serial(in),
             spice_header_get_msg_type(in->header, c->use_mini_header),
             spice_header_get_msg_size(in->header, c->use_mini_header),
@@ -582,7 +581,7 @@ void spice_msg_out_hexdump(SpiceMsgOut *out, unsigned char *data, int len)
 {
     SpiceChannelPrivate *c = out->channel->priv;
 
-    fprintf(stderr, "--\n>> hdr: %s serial %" PRIu64 " type %d size %d sub-list %d\n",
+    fprintf(stderr, "--\n>> hdr: %s serial %" PRIu64 " type %u size %u sub-list %u\n",
             c->name,
             spice_header_get_out_msg_serial(out),
             spice_header_get_msg_type(out->header, c->use_mini_header),
@@ -623,7 +622,7 @@ SpiceMsgOut *spice_msg_out_new(SpiceChannel *channel, int type)
 
     g_return_val_if_fail(c != NULL, NULL);
 
-    out = g_slice_new0(SpiceMsgOut);
+    out = g_new0(SpiceMsgOut, 1);
     out->refcount = 1;
     out->channel  = channel;
     out->ro_check = msg_check_read_only(c->channel_type, type);
@@ -659,7 +658,7 @@ void spice_msg_out_unref(SpiceMsgOut *out)
     if (out->refcount > 0)
         return;
     spice_marshaller_destroy(out->marshaller);
-    g_slice_free(SpiceMsgOut, out);
+    g_free(out);
 }
 
 /* system context */
@@ -682,9 +681,9 @@ static gboolean spice_channel_idle_wakeup(gpointer user_data)
      *   5) xmit_queue_wakeup_id now says there is a wakeup pending which is
      *      false
      */
-    STATIC_MUTEX_LOCK(c->xmit_queue_lock);
+    g_mutex_lock(&c->xmit_queue_lock);
     c->xmit_queue_wakeup_id = 0;
-    STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
+    g_mutex_unlock(&c->xmit_queue_lock);
 
     spice_channel_wakeup(channel, FALSE);
 
@@ -697,12 +696,14 @@ void spice_msg_out_send(SpiceMsgOut *out)
 {
     SpiceChannelPrivate *c;
     gboolean was_empty;
+    guint32 size;
 
     g_return_if_fail(out != NULL);
     g_return_if_fail(out->channel != NULL);
     c = out->channel->priv;
+    size = spice_marshaller_get_total_size(out->marshaller);
 
-    STATIC_MUTEX_LOCK(c->xmit_queue_lock);
+    g_mutex_lock(&c->xmit_queue_lock);
     if (c->xmit_queue_blocked) {
         g_warning("message queue is blocked, dropping message");
         goto end;
@@ -710,6 +711,7 @@ void spice_msg_out_send(SpiceMsgOut *out)
 
     was_empty = g_queue_is_empty(&c->xmit_queue);
     g_queue_push_tail(&c->xmit_queue, out);
+    c->xmit_queue_size = (was_empty) ? size : c->xmit_queue_size + size;
 
     /* One wakeup is enough to empty the entire queue -> only do a wakeup
        if the queue was empty, and there isn't one pending already. */
@@ -722,7 +724,7 @@ void spice_msg_out_send(SpiceMsgOut *out)
     }
 
 end:
-    STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
+    g_mutex_unlock(&c->xmit_queue_lock);
 }
 
 /* coroutine context */
@@ -798,7 +800,7 @@ static void spice_channel_flush_wire(SpiceChannel *channel,
     }
 }
 
-#if HAVE_SASL
+#ifdef HAVE_SASL
 /*
  * Encode all buffered data, write all encrypted data out
  * to the wire
@@ -826,7 +828,7 @@ static void spice_channel_flush_sasl(SpiceChannel *channel, const void *data, si
 /* coroutine context */
 static void spice_channel_write(SpiceChannel *channel, const void *data, size_t len)
 {
-#if HAVE_SASL
+#ifdef HAVE_SASL
     SpiceChannelPrivate *c = channel->priv;
 
     if (c->sasl_conn)
@@ -854,6 +856,7 @@ static void spice_channel_write_msg(SpiceChannel *channel, SpiceMsgOut *out)
         return;
     }
 
+    spice_marshaller_flush(out->marshaller);
     msg_size = spice_marshaller_get_total_size(out->marshaller) -
                spice_header_get_header_size(channel->priv->use_mini_header);
     spice_header_set_msg_size(out->header, channel->priv->use_mini_header, msg_size);
@@ -867,6 +870,74 @@ static void spice_channel_write_msg(SpiceChannel *channel, SpiceMsgOut *out)
     spice_msg_out_unref(out);
 }
 
+#ifdef G_OS_UNIX
+static ssize_t read_fd(int fd, int *msgfd)
+{
+    struct msghdr msg = { NULL, };
+    struct iovec iov[1];
+    union {
+        struct cmsghdr cmsg;
+        char control[CMSG_SPACE(sizeof(int))];
+    } msg_control;
+    struct cmsghdr *cmsg;
+    ssize_t ret;
+    char c;
+
+    iov[0].iov_base = &c;
+    iov[0].iov_len = 1;
+
+    msg.msg_iov = iov;
+    msg.msg_iovlen = 1;
+    msg.msg_control = &msg_control;
+    msg.msg_controllen = sizeof(msg_control);
+
+    ret = recvmsg(fd, &msg, 0);
+    if (ret > 0) {
+        for (cmsg = CMSG_FIRSTHDR(&msg);
+             cmsg != NULL;
+             cmsg = CMSG_NXTHDR(&msg, cmsg)) {
+            if (cmsg->cmsg_len != CMSG_LEN(sizeof(int)) ||
+                cmsg->cmsg_level != SOL_SOCKET ||
+                cmsg->cmsg_type != SCM_RIGHTS) {
+                continue;
+            }
+
+            memcpy(msgfd, CMSG_DATA(cmsg), sizeof(int));
+            if (*msgfd < 0) {
+                continue;
+            }
+        }
+    }
+    return ret;
+}
+
+G_GNUC_INTERNAL
+gint spice_channel_unix_read_fd(SpiceChannel *channel)
+{
+    SpiceChannelPrivate *c = channel->priv;
+    gint fd = -1;
+
+    g_return_val_if_fail(g_socket_get_family(c->sock) == G_SOCKET_FAMILY_UNIX, -1);
+
+    while (1) {
+        /* g_socket_receive_message() is not convenient here because it
+         * reads all control messages, and overly complicated to deal with */
+        if (read_fd(g_socket_get_fd(c->sock), &fd) > 0) {
+            break;
+        }
+
+        if (errno == EWOULDBLOCK) {
+            g_coroutine_socket_wait(&c->coroutine, c->sock, G_IO_IN);
+        } else {
+            g_warning("failed to get fd: %s", g_strerror(errno));
+            return -1;
+        }
+    }
+
+    return fd;
+}
+#endif
+
 /*
  * Read at least 1 more byte of data straight off the wire
  * into the requested buffer.
@@ -927,7 +998,7 @@ reread:
     return ret;
 }
 
-#if HAVE_SASL
+#ifdef HAVE_SASL
 /*
  * Read at least 1 more byte of data out of the SASL decrypted
  * data buffer, into the internal read buffer
@@ -989,7 +1060,7 @@ static int spice_channel_read(SpiceChannel *channel, void *data, size_t length)
     while (len > 0) {
         if (c->has_error) return 0; /* has_error is set by disconnect(), return no error */
 
-#if HAVE_SASL
+#ifdef HAVE_SASL
         if (c->sasl_conn)
             ret = spice_channel_read_sasl(channel, data, len);
         else
@@ -1016,11 +1087,21 @@ static void spice_channel_failed_authentication(SpiceChannel *channel,
 {
     SpiceChannelPrivate *c = channel->priv;
 
-    if (c->auth_needs_username_and_password)
+    if (c->auth_needs_username && c->auth_needs_password)
         g_set_error_literal(&c->error,
                             SPICE_CLIENT_ERROR,
                             SPICE_CLIENT_ERROR_AUTH_NEEDS_PASSWORD_AND_USERNAME,
                             _("Authentication failed: password and username are required"));
+    else if (c->auth_needs_username)
+        g_set_error_literal(&c->error,
+                            SPICE_CLIENT_ERROR,
+                            SPICE_CLIENT_ERROR_AUTH_NEEDS_USERNAME,
+                            _("Authentication failed: username is required"));
+    else if (c->auth_needs_password)
+        g_set_error_literal(&c->error,
+                            SPICE_CLIENT_ERROR,
+                            SPICE_CLIENT_ERROR_AUTH_NEEDS_PASSWORD,
+                            _("Authentication failed: password is required"));
     else if (invalidPassword)
         g_set_error_literal(&c->error,
                             SPICE_CLIENT_ERROR,
@@ -1104,7 +1185,7 @@ static gboolean spice_channel_recv_auth(SpiceChannel *channel)
     }
 
     if (link_res != SPICE_LINK_ERR_OK) {
-        CHANNEL_DEBUG(channel, "link result: reply %d", link_res);
+        CHANNEL_DEBUG(channel, "link result: reply %u", link_res);
         spice_channel_failed_authentication(channel, FALSE);
         return FALSE;
     }
@@ -1128,7 +1209,7 @@ void spice_channel_up(SpiceChannel *channel)
 {
     SpiceChannelPrivate *c = channel->priv;
 
-    CHANNEL_DEBUG(channel, "channel up, state %d", c->state);
+    CHANNEL_DEBUG(channel, "channel up, state %u", c->state);
 
     if (SPICE_CHANNEL_GET_CLASS(channel)->channel_up)
         SPICE_CHANNEL_GET_CLASS(channel)->channel_up(channel);
@@ -1163,35 +1244,39 @@ static void spice_channel_send_link(SpiceChannel *channel)
         return;
     }
 
-    c->link_msg.connection_id = spice_session_get_connection_id(c->session);
+    c->link_hdr.major_version = GUINT32_TO_LE(c->link_hdr.major_version);
+    c->link_hdr.minor_version = GUINT32_TO_LE(c->link_hdr.minor_version);
+
+    c->link_msg.connection_id = GUINT32_TO_LE(spice_session_get_connection_id(c->session));
     c->link_msg.channel_type  = c->channel_type;
     c->link_msg.channel_id    = c->channel_id;
-    c->link_msg.caps_offset   = sizeof(c->link_msg);
+    c->link_msg.caps_offset   = GUINT32_TO_LE(sizeof(c->link_msg));
 
-    c->link_msg.num_common_caps = c->common_caps->len;
-    c->link_msg.num_channel_caps = c->caps->len;
-    c->link_hdr.size += (c->link_msg.num_common_caps +
-                         c->link_msg.num_channel_caps) * sizeof(uint32_t);
+    c->link_msg.num_common_caps = GUINT32_TO_LE(c->common_caps->len);
+    c->link_msg.num_channel_caps = GUINT32_TO_LE(c->caps->len);
+    c->link_hdr.size += (c->common_caps->len + c->caps->len) * sizeof(uint32_t);
 
     buffer = g_malloc0(sizeof(c->link_hdr) + c->link_hdr.size);
     p = buffer;
 
+    c->link_hdr.size = GUINT32_TO_LE(c->link_hdr.size);
+
     memcpy(p, &c->link_hdr, sizeof(c->link_hdr)); p += sizeof(c->link_hdr);
     memcpy(p, &c->link_msg, sizeof(c->link_msg)); p += sizeof(c->link_msg);
 
     for (i = 0; i < c->common_caps->len; i++) {
-        *(uint32_t *)p = g_array_index(c->common_caps, uint32_t, i);
+        *(uint32_t *)p = GUINT32_TO_LE(g_array_index(c->common_caps, uint32_t, i));
         p += sizeof(uint32_t);
     }
     for (i = 0; i < c->caps->len; i++) {
-        *(uint32_t *)p = g_array_index(c->caps, uint32_t, i);
+        *(uint32_t *)p = GUINT32_TO_LE(g_array_index(c->caps, uint32_t, i));
         p += sizeof(uint32_t);
     }
-    CHANNEL_DEBUG(channel, "channel type %d id %d num common caps %d num caps %d",
-                  c->link_msg.channel_type,
-                  c->link_msg.channel_id,
-                  c->link_msg.num_common_caps,
-                  c->link_msg.num_channel_caps);
+    CHANNEL_DEBUG(channel, "channel type %d id %d num common caps %u num caps %u",
+                  c->channel_type,
+                  c->channel_id,
+                  c->common_caps->len,
+                  c->caps->len);
     spice_channel_write(channel, buffer, p - buffer);
     g_free(buffer);
 }
@@ -1213,13 +1298,19 @@ static gboolean spice_channel_recv_link_hdr(SpiceChannel *channel)
         goto error;
     }
 
-    CHANNEL_DEBUG(channel, "Peer version: %d:%d", c->peer_hdr.major_version, c->peer_hdr.minor_version);
+    CHANNEL_DEBUG(channel, "Peer version: %u:%u",
+                  GUINT32_FROM_LE(c->peer_hdr.major_version),
+                  GUINT32_FROM_LE(c->peer_hdr.minor_version));
     if (c->peer_hdr.major_version != c->link_hdr.major_version) {
-        g_warning("major mismatch (got %d, expected %d)",
+        g_warning("major mismatch (got %u, expected %u)",
                   c->peer_hdr.major_version, c->link_hdr.major_version);
         goto error;
     }
 
+    c->peer_hdr.major_version = GUINT32_FROM_LE(c->peer_hdr.major_version);
+    c->peer_hdr.minor_version = GUINT32_FROM_LE(c->peer_hdr.minor_version);
+    c->peer_hdr.size = GUINT32_FROM_LE(c->peer_hdr.size);
+
     c->peer_msg = g_malloc0(c->peer_hdr.size);
     if (c->peer_msg == NULL) {
         g_warning("invalid peer header size: %u", c->peer_hdr.size);
@@ -1243,7 +1334,7 @@ error:
     return FALSE;
 }
 
-#if HAVE_SASL
+#ifdef HAVE_SASL
 /*
  * NB, keep in sync with similar method in spice/server/reds.c
  */
@@ -1284,12 +1375,22 @@ spice_channel_gather_sasl_credentials(SpiceChannel *channel,
         switch (interact[ninteract].id) {
         case SASL_CB_AUTHNAME:
         case SASL_CB_USER:
-            c->auth_needs_username_and_password = TRUE;
-            if (spice_session_get_username(c->session) == NULL)
-                return FALSE;
+            c->auth_needs_username = TRUE;
+            break;
+        case SASL_CB_PASS:
+            c->auth_needs_password = TRUE;
+            break;
+        }
+    }
 
-            interact[ninteract].result =  spice_session_get_username(c->session);
-            interact[ninteract].len = strlen(interact[ninteract].result);
+    for (ninteract = 0 ; interact[ninteract].id != 0 ; ninteract++) {
+        switch (interact[ninteract].id) {
+        case SASL_CB_AUTHNAME:
+        case SASL_CB_USER:
+            if (spice_session_get_username(c->session) != NULL) {
+                interact[ninteract].result =  spice_session_get_username(c->session);
+                interact[ninteract].len = strlen(interact[ninteract].result);
+            }
             break;
 
         case SASL_CB_PASS:
@@ -1465,7 +1566,7 @@ static gboolean spice_channel_perform_auth_sasl(SpiceChannel *channel)
     if (c->has_error)
         goto error;
     if (len > SASL_MAX_MECHLIST_LEN) {
-        g_critical("mechlistlen %d too long", len);
+        g_critical("mechlistlen %u too long", len);
         goto error;
     }
 
@@ -1500,11 +1601,11 @@ restart:
         goto restart;
     }
 
-    CHANNEL_DEBUG(channel, "Server start negotiation with mech %s. Data %d bytes %p '%s'",
+    CHANNEL_DEBUG(channel, "Server start negotiation with mech %s. Data %u bytes %p '%s'",
                   mechname, clientoutlen, clientout, clientout);
 
     if (clientoutlen > SASL_MAX_DATA_LEN) {
-        g_critical("SASL negotiation data too long: %d bytes",
+        g_critical("SASL negotiation data too long: %u bytes",
                    clientoutlen);
         goto error;
     }
@@ -1533,7 +1634,7 @@ restart:
     if (c->has_error)
         goto error;
     if (len > SASL_MAX_DATA_LEN) {
-        g_critical("SASL negotiation data too long: %d bytes",
+        g_critical("SASL negotiation data too long: %u bytes",
                    len);
         goto error;
     }
@@ -1551,7 +1652,7 @@ restart:
     if (c->has_error)
         goto error;
 
-    CHANNEL_DEBUG(channel, "Client start result complete: %d. Data %d bytes %p '%s'",
+    CHANNEL_DEBUG(channel, "Client start result complete: %d. Data %u bytes %p '%s'",
                 complete, len, serverin, serverin);
 
     /* Loop-the-loop...
@@ -1584,12 +1685,9 @@ restart:
             goto restep;
         }
 
-        if (serverin) {
-            g_free(serverin);
-            serverin = NULL;
-        }
+        g_clear_pointer(&serverin, g_free);
 
-        CHANNEL_DEBUG(channel, "Client step result %d. Data %d bytes %p '%s'", err, clientoutlen, clientout, clientout);
+        CHANNEL_DEBUG(channel, "Client step result %d. Data %u bytes %p '%s'", err, clientoutlen, clientout, clientout);
 
         /* Previous server call showed completion & we're now locally complete too */
         if (complete && err == SASL_OK)
@@ -1610,13 +1708,13 @@ restart:
         if (c->has_error)
             goto error;
 
-        CHANNEL_DEBUG(channel, "Server step with %d bytes %p", clientoutlen, clientout);
+        CHANNEL_DEBUG(channel, "Server step with %u bytes %p", clientoutlen, clientout);
 
         spice_channel_read(channel, &len, sizeof(guint32));
         if (c->has_error)
             goto error;
         if (len > SASL_MAX_DATA_LEN) {
-            g_critical("SASL negotiation data too long: %d bytes", len);
+            g_critical("SASL negotiation data too long: %u bytes", len);
             goto error;
         }
 
@@ -1634,13 +1732,12 @@ restart:
         if (c->has_error)
             goto error;
 
-        CHANNEL_DEBUG(channel, "Client step result complete: %d. Data %d bytes %p '%s'",
+        CHANNEL_DEBUG(channel, "Client step result complete: %d. Data %u bytes %p '%s'",
                     complete, len, serverin, serverin);
 
         /* This server call shows complete, and earlier client step was OK */
         if (complete) {
-            g_free(serverin);
-            serverin = NULL;
+            g_clear_pointer(&serverin, g_free);
             if (err == SASL_CONTINUE) /* something went wrong */
                 goto complete;
             break;
@@ -1656,9 +1753,9 @@ restart:
             goto error;
         }
         ssf = *(const int *)val;
-        CHANNEL_DEBUG(channel, "SASL SSF value %d", ssf);
+        CHANNEL_DEBUG(channel, "SASL SSF value %u", ssf);
         if (ssf < 56) { /* 56 == DES level, good for Kerberos */
-            g_critical("negotiation SSF %d was not strong enough", ssf);
+            g_critical("negotiation SSF %u was not strong enough", ssf);
             goto error;
         }
     }
@@ -1697,7 +1794,7 @@ static gboolean spice_channel_recv_link_msg(SpiceChannel *channel)
 {
     SpiceChannelPrivate *c;
     int rc, num_caps, i;
-    uint32_t *caps;
+    uint32_t *caps, num_channel_caps, num_common_caps;
     SpiceChannelEvent event = SPICE_CHANNEL_ERROR_LINK;
 
     g_return_val_if_fail(channel != NULL, FALSE);
@@ -1709,7 +1806,7 @@ static gboolean spice_channel_recv_link_msg(SpiceChannel *channel)
                             c->peer_hdr.size - c->peer_pos);
     c->peer_pos += rc;
     if (c->peer_pos != c->peer_hdr.size) {
-        g_critical("%s: %s: incomplete link reply (%d/%d)",
+        g_critical("%s: %s: incomplete link reply (%d/%u)",
                   c->name, __FUNCTION__, rc, c->peer_hdr.size);
         goto error;
     }
@@ -1723,29 +1820,32 @@ static gboolean spice_channel_recv_link_msg(SpiceChannel *channel)
         c->tls = TRUE;
         return FALSE;
     default:
-        g_warning("%s: %s: unhandled error %d",
+        g_warning("%s: %s: unhandled error %u",
                 c->name, __FUNCTION__, c->peer_msg->error);
         goto error;
     }
 
-    num_caps = c->peer_msg->num_channel_caps + c->peer_msg->num_common_caps;
+    num_channel_caps = GUINT32_FROM_LE(c->peer_msg->num_channel_caps);
+    num_common_caps = GUINT32_FROM_LE(c->peer_msg->num_common_caps);
+
+    num_caps = num_channel_caps + num_common_caps;
     CHANNEL_DEBUG(channel, "%s: %d caps", __FUNCTION__, num_caps);
 
     /* see original spice/client code: */
     /* g_return_if_fail(c->peer_msg + c->peer_msg->caps_offset * sizeof(uint32_t) > c->peer_msg + c->peer_hdr.size); */
 
-    caps = (uint32_t *)((uint8_t *)c->peer_msg + c->peer_msg->caps_offset);
+    caps = (uint32_t *)((uint8_t *)c->peer_msg + GUINT32_FROM_LE(c->peer_msg->caps_offset));
 
-    g_array_set_size(c->remote_common_caps, c->peer_msg->num_common_caps);
-    for (i = 0; i < c->peer_msg->num_common_caps; i++, caps++) {
-        g_array_index(c->remote_common_caps, uint32_t, i) = *caps;
-        CHANNEL_DEBUG(channel, "got common caps %u:0x%X", i, *caps);
+    g_array_set_size(c->remote_common_caps, num_common_caps);
+    for (i = 0; i < num_common_caps; i++, caps++) {
+        g_array_index(c->remote_common_caps, uint32_t, i) = GUINT32_FROM_LE(*caps);
+        CHANNEL_DEBUG(channel, "got common caps %d:0x%X", i, GUINT32_FROM_LE(*caps));
     }
 
-    g_array_set_size(c->remote_caps, c->peer_msg->num_channel_caps);
-    for (i = 0; i < c->peer_msg->num_channel_caps; i++, caps++) {
-        g_array_index(c->remote_caps, uint32_t, i) = *caps;
-        CHANNEL_DEBUG(channel, "got channel caps %u:0x%X", i, *caps);
+    g_array_set_size(c->remote_caps, num_channel_caps);
+    for (i = 0; i < num_channel_caps; i++, caps++) {
+        g_array_index(c->remote_caps, uint32_t, i) = GUINT32_FROM_LE(*caps);
+        CHANNEL_DEBUG(channel, "got channel caps %d:0x%X", i, GUINT32_FROM_LE(*caps));
     }
 
     if (!spice_channel_test_common_capability(channel,
@@ -1756,7 +1856,7 @@ static gboolean spice_channel_recv_link_msg(SpiceChannel *channel)
     } else {
         SpiceLinkAuthMechanism auth = { 0, };
 
-#if HAVE_SASL
+#ifdef HAVE_SASL
         if (spice_channel_test_common_capability(channel, SPICE_COMMON_CAP_AUTH_SASL)) {
             CHANNEL_DEBUG(channel, "Choosing SASL mechanism");
             auth.auth_mechanism = SPICE_COMMON_CAP_AUTH_SASL;
@@ -1790,7 +1890,10 @@ error:
 G_GNUC_INTERNAL
 void spice_channel_wakeup(SpiceChannel *channel, gboolean cancel)
 {
-    GCoroutine *c = &channel->priv->coroutine;
+    GCoroutine *c;
+
+    g_return_if_fail(SPICE_IS_CHANNEL(channel));
+    c = &channel->priv->coroutine;
 
     if (cancel)
         g_coroutine_condition_cancel(c);
@@ -2009,7 +2112,7 @@ SpiceChannel *spice_channel_new(SpiceSession *s, int type, int id)
     case SPICE_CHANNEL_PLAYBACK:
     case SPICE_CHANNEL_RECORD: {
         if (!spice_session_get_audio_enabled(s)) {
-            g_debug("audio channel is disabled, not creating it");
+            SPICE_DEBUG("audio channel is disabled, not creating it");
             return NULL;
         }
         gtype = type == SPICE_CHANNEL_RECORD ?
@@ -2019,7 +2122,7 @@ SpiceChannel *spice_channel_new(SpiceSession *s, int type, int id)
 #ifdef USE_SMARTCARD
     case SPICE_CHANNEL_SMARTCARD: {
         if (!spice_session_get_smartcard_enabled(s)) {
-            g_debug("smartcard channel is disabled, not creating it");
+            SPICE_DEBUG("smartcard channel is disabled, not creating it");
             return NULL;
         }
         gtype = SPICE_TYPE_SMARTCARD_CHANNEL;
@@ -2029,7 +2132,7 @@ SpiceChannel *spice_channel_new(SpiceSession *s, int type, int id)
 #ifdef USE_USBREDIR
     case SPICE_CHANNEL_USBREDIR: {
         if (!spice_session_get_usbredir_enabled(s)) {
-            g_debug("usbredir channel is disabled, not creating it");
+            SPICE_DEBUG("usbredir channel is disabled, not creating it");
             return NULL;
         }
         gtype = SPICE_TYPE_USBREDIR_CHANNEL;
@@ -2046,7 +2149,7 @@ SpiceChannel *spice_channel_new(SpiceSession *s, int type, int id)
         gtype = SPICE_TYPE_PORT_CHANNEL;
         break;
     default:
-        g_debug("unsupported channel kind: %s: %d",
+        SPICE_DEBUG("unsupported channel kind: %s: %d",
                 spice_channel_type_to_string(type), type);
         return NULL;
     }
@@ -2085,9 +2188,7 @@ static void spice_channel_flushed(SpiceChannel *channel, gboolean success)
     GSList *l;
 
     for (l = c->flushing; l != NULL; l = l->next) {
-        GSimpleAsyncResult *result = G_SIMPLE_ASYNC_RESULT(l->data);
-        g_simple_async_result_set_op_res_gboolean(result, success);
-        g_simple_async_result_complete_in_idle(result);
+        g_task_return_boolean(G_TASK(l->data), success);
     }
 
     g_slist_free_full(c->flushing, g_object_unref);
@@ -2101,11 +2202,14 @@ static void spice_channel_iterate_write(SpiceChannel *channel)
     SpiceMsgOut *out;
 
     do {
-        STATIC_MUTEX_LOCK(c->xmit_queue_lock);
+        g_mutex_lock(&c->xmit_queue_lock);
         out = g_queue_pop_head(&c->xmit_queue);
-        STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
-        if (out)
+        g_mutex_unlock(&c->xmit_queue_lock);
+        if (out) {
+            guint32 size = spice_marshaller_get_total_size(out->marshaller);
+            c->xmit_queue_size = (c->xmit_queue_size < size) ? 0 : c->xmit_queue_size - size;
             spice_channel_write_msg(channel, out);
+        }
     } while (out);
 
     spice_channel_flushed(channel, TRUE);
@@ -2125,7 +2229,7 @@ static void spice_channel_iterate_read(SpiceChannel *channel)
     ) { do
             spice_channel_recv_msg(channel,
                                    (handler_msg_in)SPICE_CHANNEL_GET_CLASS(channel)->handle_msg, NULL);
-#if HAVE_SASL
+#ifdef HAVE_SASL
             /* flush the sasl buffer too */
         while (c->sasl_decoded != NULL);
 #else
@@ -2196,6 +2300,7 @@ static gboolean spice_channel_delayed_unref(gpointer data)
     SpiceChannel *channel = SPICE_CHANNEL(data);
     SpiceChannelPrivate *c = channel->priv;
     gboolean was_ready = c->state == SPICE_CHANNEL_STATE_READY;
+    SpiceSession *session;
 
     CHANNEL_DEBUG(channel, "Delayed unref channel %p", channel);
 
@@ -2203,6 +2308,13 @@ static gboolean spice_channel_delayed_unref(gpointer data)
 
     c->state = SPICE_CHANNEL_STATE_UNCONNECTED;
 
+    session = spice_channel_get_session(channel);
+    if (session && spice_session_is_for_migration(session)) {
+        /* error during migration - abort migration */
+        spice_session_abort_migration(session);
+        return FALSE;
+    }
+
     if (c->event != SPICE_CHANNEL_NONE) {
         g_coroutine_signal_emit(channel, signals[SPICE_CHANNEL_EVENT], 0, c->event);
         c->event = SPICE_CHANNEL_NONE;
@@ -2570,7 +2682,7 @@ gboolean spice_channel_open_fd(SpiceChannel *channel, int fd)
 
     c = channel->priv;
     if (c->state > SPICE_CHANNEL_STATE_CONNECTING) {
-        g_warning("Invalid channel_connect state: %d", c->state);
+        g_warning("Invalid channel_connect state: %u", c->state);
         return true;
     }
 
@@ -2590,7 +2702,7 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
         c->connect_delayed_id = 0;
     }
 
-#if HAVE_SASL
+#ifdef HAVE_SASL
     if (c->sasl_conn) {
         sasl_dispose(&c->sasl_conn);
         c->sasl_conn = NULL;
@@ -2598,35 +2710,22 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
     }
 #endif
 
-    spice_openssl_verify_free(c->sslverify);
-    c->sslverify = NULL;
-
-    if (c->ssl) {
-        SSL_free(c->ssl);
-        c->ssl = NULL;
-    }
-
-    if (c->ctx) {
-        SSL_CTX_free(c->ctx);
-        c->ctx = NULL;
-    }
-
-    if (c->conn) {
-        g_object_unref(c->conn);
-        c->conn = NULL;
-    }
+    g_clear_pointer(&c->sslverify, spice_openssl_verify_free);
+    g_clear_pointer(&c->ssl, SSL_free);
+    g_clear_pointer(&c->ctx, SSL_CTX_free);
 
+    g_clear_object(&c->conn);
     g_clear_object(&c->sock);
 
     c->fd = -1;
 
-    c->auth_needs_username_and_password = FALSE;
+    c->auth_needs_username = FALSE;
+    c->auth_needs_password = FALSE;
 
-    g_free(c->peer_msg);
-    c->peer_msg = NULL;
+    g_clear_pointer(&c->peer_msg, g_free);
     c->peer_pos = 0;
 
-    STATIC_MUTEX_LOCK(c->xmit_queue_lock);
+    g_mutex_lock(&c->xmit_queue_lock);
     c->xmit_queue_blocked = TRUE; /* Disallow queuing new messages */
     gboolean was_empty = g_queue_is_empty(&c->xmit_queue);
     g_queue_foreach(&c->xmit_queue, (GFunc)spice_msg_out_unref, NULL);
@@ -2635,7 +2734,7 @@ static void channel_reset(SpiceChannel *channel, gboolean migrating)
         g_source_remove(c->xmit_queue_wakeup_id);
         c->xmit_queue_wakeup_id = 0;
     }
-    STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
+    g_mutex_unlock(&c->xmit_queue_lock);
     spice_channel_flushed(channel, was_empty);
 
     g_array_set_size(c->remote_common_caps, 0);
@@ -2672,7 +2771,7 @@ void spice_channel_disconnect(SpiceChannel *channel, SpiceChannelEvent reason)
 {
     SpiceChannelPrivate *c;
 
-    CHANNEL_DEBUG(channel, "channel disconnect %d", reason);
+    CHANNEL_DEBUG(channel, "channel disconnect %u", reason);
 
     g_return_if_fail(SPICE_IS_CHANNEL(channel));
     g_return_if_fail(channel->priv != NULL);
@@ -2710,7 +2809,7 @@ static gboolean test_capability(GArray *caps, guint32 cap)
     c = g_array_index(caps, guint32, word_index);
     ret = (c & (1 << (cap % 32))) != 0;
 
-    SPICE_DEBUG("test cap %d in 0x%X: %s", cap, c, ret ? "yes" : "no");
+    SPICE_DEBUG("test cap %u in 0x%X: %s", cap, c, ret ? "yes" : "no");
     return ret;
 }
 
@@ -2813,6 +2912,17 @@ enum spice_channel_state spice_channel_get_state(SpiceChannel *channel)
     return channel->priv->state;
 }
 
+G_GNUC_INTERNAL
+guint64 spice_channel_get_queue_size (SpiceChannel *channel)
+{
+    guint64 size;
+    SpiceChannelPrivate *c = channel->priv;
+    g_mutex_lock(&c->xmit_queue_lock);
+    size = c->xmit_queue_size;
+    g_mutex_unlock(&c->xmit_queue_lock);
+    return size;
+}
+
 G_GNUC_INTERNAL
 void spice_channel_swap(SpiceChannel *channel, SpiceChannel *swap, gboolean swap_msgs)
 {
@@ -2852,7 +2962,7 @@ void spice_channel_swap(SpiceChannel *channel, SpiceChannel *swap, gboolean swap
     SWAP(common_caps);
     SWAP(remote_caps);
     SWAP(remote_common_caps);
-#if HAVE_SASL
+#ifdef HAVE_SASL
     SWAP(sasl_conn);
     SWAP(sasl_decoded);
     SWAP(sasl_decoded_length);
@@ -2916,7 +3026,7 @@ static void spice_channel_send_migration_handshake(SpiceChannel *channel)
 void spice_channel_flush_async(SpiceChannel *self, GCancellable *cancellable,
                                GAsyncReadyCallback callback, gpointer user_data)
 {
-    GSimpleAsyncResult *simple;
+    GTask *task;
     SpiceChannelPrivate *c;
     gboolean was_empty;
 
@@ -2924,26 +3034,25 @@ void spice_channel_flush_async(SpiceChannel *self, GCancellable *cancellable,
     c = self->priv;
 
     if (c->state != SPICE_CHANNEL_STATE_READY) {
-        g_simple_async_report_error_in_idle(G_OBJECT(self), callback, user_data,
+        g_task_report_new_error(self, callback, user_data,
+            spice_channel_flush_async,
             SPICE_CLIENT_ERROR, SPICE_CLIENT_ERROR_FAILED,
             "The channel is not ready yet");
         return;
     }
 
-    simple = g_simple_async_result_new(G_OBJECT(self), callback, user_data,
-                                       spice_channel_flush_async);
+    task = g_task_new(self, cancellable, callback, user_data);
 
-    STATIC_MUTEX_LOCK(c->xmit_queue_lock);
+    g_mutex_lock(&c->xmit_queue_lock);
     was_empty = g_queue_is_empty(&c->xmit_queue);
-    STATIC_MUTEX_UNLOCK(c->xmit_queue_lock);
+    g_mutex_unlock(&c->xmit_queue_lock);
     if (was_empty) {
-        g_simple_async_result_set_op_res_gboolean(simple, TRUE);
-        g_simple_async_result_complete_in_idle(simple);
-        g_object_unref(simple);
+        g_task_return_boolean(task, TRUE);
+        g_object_unref(task);
         return;
     }
 
-    c->flushing = g_slist_append(c->flushing, simple);
+    c->flushing = g_slist_append(c->flushing, task);
 }
 
 /**
@@ -2961,19 +3070,16 @@ void spice_channel_flush_async(SpiceChannel *self, GCancellable *cancellable,
 gboolean spice_channel_flush_finish(SpiceChannel *self, GAsyncResult *result,
                                     GError **error)
 {
-    GSimpleAsyncResult *simple;
+    GTask *task;
 
     g_return_val_if_fail(SPICE_IS_CHANNEL(self), FALSE);
     g_return_val_if_fail(result != NULL, FALSE);
 
-    simple = (GSimpleAsyncResult *)result;
+    task = G_TASK(result);
 
-    if (g_simple_async_result_propagate_error(simple, error))
-        return -1;
-
-    g_return_val_if_fail(g_simple_async_result_is_valid(result, G_OBJECT(self),
-                                                        spice_channel_flush_async), FALSE);
+    g_return_val_if_fail(g_task_is_valid(task, self),
+                         FALSE);
 
     CHANNEL_DEBUG(self, "flushed finished!");
-    return g_simple_async_result_get_op_res_gboolean(simple);
+    return g_task_propagate_boolean(task, error);
 }
