@@ -19,7 +19,7 @@
 
 #include <glib.h>
 
-#if HAVE_X11_XKBLIB_H
+#ifdef HAVE_X11_XKBLIB_H
 #include <X11/XKBlib.h>
 #include <gdk/gdkx.h>
 #endif
@@ -38,7 +38,6 @@
 #include <gtk/gtk.h>
 #include <spice/vd_agent.h>
 #include "desktop-integration.h"
-#include "gtk-compat.h"
 #include "spice-common.h"
 #include "spice-gtk-session.h"
 #include "spice-gtk-session-priv.h"
@@ -64,6 +63,9 @@ struct _SpiceGtkSessionPrivate {
     gboolean                auto_usbredir_enable;
     int                     auto_usbredir_reqs;
     gboolean                pointer_grabbed;
+    gboolean                keyboard_has_focus;
+    gboolean                mouse_has_pointer;
+    gboolean                sync_modifiers;
 };
 
 /**
@@ -73,7 +75,7 @@ struct _SpiceGtkSessionPrivate {
  * @section_id:
  * @see_also: #SpiceSession, and the GTK widget #SpiceDisplay
  * @stability: Stable
- * @include: spice-gtk-session.h
+ * @include: spice-client-gtk.h
  *
  * The #SpiceGtkSession class is the spice-client-gtk counter part of
  * #SpiceSession. It contains functionality which should be handled per
@@ -117,6 +119,7 @@ enum {
     PROP_AUTO_CLIPBOARD,
     PROP_AUTO_USBREDIR,
     PROP_POINTER_GRABBED,
+    PROP_SYNC_MODIFIERS,
 };
 
 static guint32 get_keyboard_lock_modifiers(void)
@@ -137,7 +140,7 @@ static guint32 get_keyboard_lock_modifiers(void)
         modifiers |= SPICE_INPUTS_SCROLL_LOCK;
     }
 #else
-#if HAVE_X11_XKBLIB_H
+#ifdef HAVE_X11_XKBLIB_H
     Display *x_display = NULL;
     XKeyboardState keyboard_state;
 
@@ -180,10 +183,15 @@ static void spice_gtk_session_sync_keyboard_modifiers_for_channel(SpiceGtkSessio
                                                                   SpiceInputsChannel* inputs,
                                                                   gboolean force)
 {
-    gint guest_modifiers = 0, client_modifiers = 0;
+    guint32 guest_modifiers = 0, client_modifiers = 0;
 
     g_return_if_fail(SPICE_IS_INPUTS_CHANNEL(inputs));
 
+    if (SPICE_IS_GTK_SESSION(self) && !self->priv->sync_modifiers) {
+        SPICE_DEBUG("Syncing modifiers is disabled");
+        return;
+    }
+
     g_object_get(inputs, "key-modifiers", &guest_modifiers, NULL);
     client_modifiers = get_keyboard_lock_modifiers();
 
@@ -302,8 +310,7 @@ static void spice_gtk_session_finalize(GObject *gobject)
 
     /* release stuff */
     for (i = 0; i < CLIPBOARD_LAST; ++i) {
-        g_free(s->clip_targets[i]);
-        s->clip_targets[i] = NULL;
+        g_clear_pointer(&s->clip_targets[i], g_free);
     }
 
     /* Chain up to the parent class */
@@ -332,6 +339,9 @@ static void spice_gtk_session_get_property(GObject    *gobject,
     case PROP_POINTER_GRABBED:
         g_value_set_boolean(value, s->pointer_grabbed);
         break;
+    case PROP_SYNC_MODIFIERS:
+        g_value_set_boolean(value, s->sync_modifiers);
+        break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
 	break;
@@ -379,6 +389,9 @@ static void spice_gtk_session_set_property(GObject      *gobject,
         }
         break;
     }
+    case PROP_SYNC_MODIFIERS:
+        s->sync_modifiers = g_value_get_boolean(value);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
         break;
@@ -467,6 +480,23 @@ static void spice_gtk_session_class_init(SpiceGtkSessionClass *klass)
                               G_PARAM_READABLE |
                               G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceGtkSession:sync-modifiers:
+     *
+     * Automatically sync modifiers (Caps, Num and Scroll locks) with the guest.
+     *
+     * Since: 0.32
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_SYNC_MODIFIERS,
+         g_param_spec_boolean("sync-modifiers",
+                              "Sync modifiers",
+                              "Automatically sync modifiers",
+                              TRUE,
+                              G_PARAM_READWRITE |
+                              G_PARAM_CONSTRUCT |
+                              G_PARAM_STATIC_STRINGS));
+
     g_type_class_add_private(klass, sizeof(SpiceGtkSessionPrivate));
 }
 
@@ -481,7 +511,7 @@ static GtkClipboard* get_clipboard_from_selection(SpiceGtkSessionPrivate *s,
     } else if (selection == VD_AGENT_CLIPBOARD_SELECTION_PRIMARY) {
         return s->clipboard_primary;
     } else {
-        g_warning("Unhandled clipboard selection: %d", selection);
+        g_warning("Unhandled clipboard selection: %u", selection);
         return NULL;
     }
 }
@@ -791,8 +821,7 @@ static void clipboard_get(GtkClipboard *clipboard,
     gdk_threads_enter();
 
 cleanup:
-    g_main_loop_unref(ri.loop);
-    ri.loop = NULL;
+    g_clear_pointer(&ri.loop, g_main_loop_unref);
     g_signal_handler_disconnect(s->main, clipboard_handler);
     g_signal_handler_disconnect(s->main, agent_handler);
 }
@@ -835,7 +864,7 @@ static gboolean clipboard_grab(SpiceMainChannel *main, guint selection,
             }
         }
         if (!found) {
-            g_warning("clipboard: couldn't find a matching type for: %d",
+            g_warning("clipboard: couldn't find a matching type for: %u",
                       types[n]);
         }
     }
@@ -1132,15 +1161,15 @@ SpiceGtkSession *spice_gtk_session_get(SpiceSession *session)
     g_return_val_if_fail(SPICE_IS_SESSION(session), NULL);
 
     SpiceGtkSession *self;
-    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
+    static GMutex mutex;
 
-    g_static_mutex_lock(&mutex);
+    g_mutex_lock(&mutex);
     self = g_object_get_data(G_OBJECT(session), "spice-gtk-session");
     if (self == NULL) {
         self = g_object_new(SPICE_TYPE_GTK_SESSION, "session", session, NULL);
         g_object_set_data_full(G_OBJECT(session), "spice-gtk-session", self, g_object_unref);
     }
-    g_static_mutex_unlock(&mutex);
+    g_mutex_unlock(&mutex);
 
     return SPICE_GTK_SESSION(self);
 }
@@ -1227,3 +1256,36 @@ gboolean spice_gtk_session_get_pointer_grabbed(SpiceGtkSession *self)
 
     return self->priv->pointer_grabbed;
 }
+
+G_GNUC_INTERNAL
+void spice_gtk_session_set_keyboard_has_focus(SpiceGtkSession *self,
+                                                gboolean keyboard_has_focus)
+{
+    g_return_if_fail(SPICE_IS_GTK_SESSION(self));
+
+    self->priv->keyboard_has_focus = keyboard_has_focus;
+}
+
+G_GNUC_INTERNAL
+void spice_gtk_session_set_mouse_has_pointer(SpiceGtkSession *self,
+                                                gboolean mouse_has_pointer)
+{
+   g_return_if_fail(SPICE_IS_GTK_SESSION(self));
+   self->priv->mouse_has_pointer = mouse_has_pointer;
+}
+
+G_GNUC_INTERNAL
+gboolean spice_gtk_session_get_keyboard_has_focus(SpiceGtkSession *self)
+{
+    g_return_val_if_fail(SPICE_IS_GTK_SESSION(self), FALSE);
+
+    return self->priv->keyboard_has_focus;
+}
+
+G_GNUC_INTERNAL
+gboolean spice_gtk_session_get_mouse_has_pointer(SpiceGtkSession *self)
+{
+    g_return_val_if_fail(SPICE_IS_GTK_SESSION(self), FALSE);
+
+    return self->priv->mouse_has_pointer;
+}
