@@ -30,7 +30,6 @@
 #include "spice-util-priv.h"
 #include "spice-session-priv.h"
 #include "gio-coroutine.h"
-#include "glib-compat.h"
 #include "wocky-http-proxy.h"
 #include "spice-uri-priv.h"
 #include "channel-playback-priv.h"
@@ -134,7 +133,7 @@ struct _SpiceSessionPrivate {
  * @section_id:
  * @see_also: #SpiceChannel, and the GTK widget #SpiceDisplay
  * @stability: Stable
- * @include: spice-session.h
+ * @include: spice-client.h
  *
  * The #SpiceSession class handles all the #SpiceChannel connections.
  * It's also the class that contains connections informations, such as
@@ -318,8 +317,7 @@ session_disconnect(SpiceSession *self, gboolean keep_main)
 
     s->connection_id = 0;
 
-    g_free(s->name);
-    s->name = NULL;
+    g_clear_pointer(&s->name, g_free);
     memset(s->uuid, 0, sizeof(s->uuid));
 
     spice_session_abort_migration(self);
@@ -372,7 +370,7 @@ spice_session_finalize(GObject *gobject)
     g_strfreev(s->secure_channels);
     g_free(s->shared_dir);
 
-    g_clear_pointer(&s->images, cache_unref);
+    g_clear_pointer(&s->images, cache_free);
     glz_decoder_window_destroy(s->glz_window);
 
     g_clear_pointer(&s->pubkey, g_byte_array_unref);
@@ -418,7 +416,6 @@ static int spice_parse_uri(SpiceSession *session, const char *original_uri)
     SpiceSessionPrivate *s = session->priv;
     gchar *host = NULL, *port = NULL, *tls_port = NULL, *uri = NULL, *username = NULL, *password = NULL;
     gchar *path = NULL;
-    gchar *unescaped_path = NULL;
     gchar *authority = NULL;
     gchar *query = NULL;
     gchar *tmp = NULL;
@@ -479,7 +476,7 @@ static int spice_parse_uri(SpiceSession *session, const char *original_uri)
         }
         tmp[0] = '\0';
         tmp++;
-        host = g_strdup(authority + 1);
+        host = g_strdup_printf("[%s]", authority + 1);
         if (tmp[0] == ':')
             port = g_strdup(tmp + 1);
     } else {
@@ -497,7 +494,6 @@ static int spice_parse_uri(SpiceSession *session, const char *original_uri)
         g_warning("Unexpected path data '%s' for URI '%s'", path, uri);
         /* don't fail, just ignore */
     }
-    unescaped_path = g_uri_unescape_string(path, NULL);
     path = NULL;
 
     while (query && query[0] != '\0') {
@@ -506,13 +502,13 @@ static int spice_parse_uri(SpiceSession *session, const char *original_uri)
 
         int len;
         if (sscanf(query, "%31[-a-zA-Z0-9]=%n", key, &len) != 1) {
-            spice_warning("Failed to parse key in URI '%s'", query);
+            g_warning("Failed to parse key in URI '%s'", query);
             goto fail;
         }
 
         query += len;
         if (*query == '\0') {
-            spice_warning ("key '%s' without value", key);
+            SPICE_DEBUG("key '%s' without value", key);
             break;
         } else if (*query == ';' || *query == '&') {
             /* another argument */
@@ -521,7 +517,7 @@ static int spice_parse_uri(SpiceSession *session, const char *original_uri)
         }
 
         if (sscanf(query, "%127[^;&]%n", value, &len) != 1) {
-            spice_warning("Failed to parse value of key '%s' in URI '%s'", key, query);
+            g_warning("Failed to parse value of key '%s' in URI '%s'", key, query);
             goto fail;
         }
 
@@ -557,8 +553,6 @@ static int spice_parse_uri(SpiceSession *session, const char *original_uri)
 
 end:
     /* parsed ok -> apply */
-    g_free(uri);
-    g_free(unescaped_path);
     g_free(s->unix_path);
     g_free(s->host);
     g_free(s->port);
@@ -566,6 +560,7 @@ end:
     g_free(s->username);
     g_free(s->password);
     s->unix_path = g_strdup(path);
+    g_free(uri);
     s->host = host;
     s->port = port;
     s->tls_port = tls_port;
@@ -575,7 +570,6 @@ end:
 
 fail:
     g_free(uri);
-    g_free(unescaped_path);
     g_free(host);
     g_free(port);
     g_free(tls_port);
@@ -1696,7 +1690,7 @@ void spice_session_start_migrating(SpiceSession *session,
 
     g_warn_if_fail(ring_get_length(&s->channels) == ring_get_length(&m->channels));
 
-    SPICE_DEBUG("migration channels left:%d (in migration:%d)",
+    SPICE_DEBUG("migration channels left:%u (in migration:%u)",
                 ring_get_length(&s->channels), ring_get_length(&m->channels));
     s->migration_left = spice_session_get_channels(session);
 }
@@ -1763,11 +1757,9 @@ void spice_session_abort_migration(SpiceSession *session)
     }
 
 end:
-    g_list_free(s->migration_left);
-    s->migration_left = NULL;
+    g_clear_pointer(&s->migration_left, g_list_free);
     session_disconnect(s->migration, FALSE);
-    g_object_unref(s->migration);
-    s->migration = NULL;
+    g_clear_pointer(&s->migration, g_object_unref);
 
     s->migrate_wait_init = FALSE;
     if (s->after_main_init) {
@@ -1806,8 +1798,7 @@ void spice_session_channel_migrate(SpiceSession *session, SpiceChannel *channel)
     if (g_list_length(s->migration_left) == 0) {
         CHANNEL_DEBUG(channel, "migration: all channel migrated, success");
         session_disconnect(s->migration, FALSE);
-        g_object_unref(s->migration);
-        s->migration = NULL;
+        g_clear_pointer(&s->migration, g_object_unref);
         spice_session_set_migration_state(session, SPICE_SESSION_MIGRATION_NONE);
     }
 }
@@ -1936,7 +1927,7 @@ void spice_session_disconnect(SpiceSession *session)
 
     s = session->priv;
 
-    SPICE_DEBUG("session: disconnecting %d", s->disconnecting);
+    SPICE_DEBUG("session: disconnecting %u", s->disconnecting);
     if (s->disconnecting != 0)
         return;
 
@@ -2178,7 +2169,7 @@ GSocketConnection* spice_session_channel_open_host(SpiceSession *session, SpiceC
     } else {
         port = *use_tls ? s->tls_port : s->port;
         if (port == NULL) {
-            g_debug("Missing port value, not attempting %s connection.",
+            SPICE_DEBUG("Missing port value, not attempting %s connection.",
                     *use_tls?"TLS":"unencrypted");
             return NULL;
         }
@@ -2642,18 +2633,18 @@ SpiceURI *spice_session_get_proxy_uri(SpiceSession *session)
  **/
 SpiceAudio *spice_audio_get(SpiceSession *session, GMainContext *context)
 {
-    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
+    static GMutex mutex;
     SpiceAudio *self;
 
     g_return_val_if_fail(SPICE_IS_SESSION(session), NULL);
 
-    g_static_mutex_lock(&mutex);
+    g_mutex_lock(&mutex);
     self = session->priv->audio_manager;
     if (self == NULL) {
         self = spice_audio_new(session, context, NULL);
         session->priv->audio_manager = self;
     }
-    g_static_mutex_unlock(&mutex);
+    g_mutex_unlock(&mutex);
 
     return self;
 }
@@ -2676,19 +2667,19 @@ SpiceUsbDeviceManager *spice_usb_device_manager_get(SpiceSession *session,
                                                     GError **err)
 {
     SpiceUsbDeviceManager *self;
-    static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
+    static GMutex mutex;
 
     g_return_val_if_fail(SPICE_IS_SESSION(session), NULL);
     g_return_val_if_fail(err == NULL || *err == NULL, NULL);
 
-    g_static_mutex_lock(&mutex);
+    g_mutex_lock(&mutex);
     self = session->priv->usb_manager;
     if (self == NULL) {
         self = g_initable_new(SPICE_TYPE_USB_DEVICE_MANAGER, NULL, err,
                               "session", session, NULL);
         session->priv->usb_manager = self;
     }
-    g_static_mutex_unlock(&mutex);
+    g_mutex_unlock(&mutex);
 
     return self;
 }
@@ -2730,7 +2721,7 @@ PhodavServer* spice_session_get_webdav_server(SpiceSession *session)
 
     const gchar *shared_dir = spice_session_get_shared_dir(session);
     if (shared_dir == NULL) {
-        g_debug("No shared dir set, not creating webdav server");
+        SPICE_DEBUG("No shared dir set, not creating webdav server");
         return NULL;
     }
 
